# -*- coding: utf-8 -*-
"""
An modification the U-Net with auxiliary decoders according to 
the CCT paper:
    Yassine Ouali, Celine Hudelot and Myriam Tami:
    Semi-Supervised Semantic Segmentation With Cross-Consistency Training. 
    CVPR 2020.
    https://arxiv.org/abs/2003.09005  
Code adapted from: https://github.com/yassouali/CCT
"""
from __future__ import print_function, division

import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np 
from torch.distributions.uniform import Uniform
from pymic.net.net2d.unet2d import Encoder, Decoder

def _l2_normalize(d):
    # Normalizing per batch axis
    d_reshaped = d.view(d.shape[0], -1, *(1 for _ in range(d.dim() - 2)))
    d /= torch.norm(d_reshaped, dim=1, keepdim=True) + 1e-8
    return d



def get_r_adv(x_list, decoder, it=1, xi=1e-1, eps=10.0):
    """
    Virtual Adversarial Training according to
    https://arxiv.org/abs/1704.03976
    """
    x_detached = [item.detach() for item in x_list]
    xe_detached = x_detached[-1]
    with torch.no_grad():
        pred = F.softmax(decoder(x_detached), dim=1)

    d = torch.rand(x_list[-1].shape).sub(0.5).to(x_list[-1].device)
    d = _l2_normalize(d)

    for _ in range(it):
        d.requires_grad_()
        x_detached[-1] = xe_detached + xi * d
        pred_hat = decoder(x_detached)
        logp_hat = F.log_softmax(pred_hat, dim=1)
        adv_distance = F.kl_div(logp_hat, pred, reduction='batchmean')
        adv_distance.backward()
        d = _l2_normalize(d.grad)
        decoder.zero_grad()

    r_adv = d * eps
    return x_list[-1] + r_adv
    

class AuxiliaryDecoder(nn.Module):
    def __init__(self, params, aux_type):
        super(AuxiliaryDecoder, self).__init__()
        self.params   = params
        self.decoder  = Decoder(params)
        self.aux_type = aux_type
        uniform_range = params.get("Uniform_range".lower(), 0.3)
        self.uni_dist = Uniform(-uniform_range, uniform_range)

    def feature_drop(self, x):
        attention  = torch.mean(x, dim=1, keepdim=True)
        max_val, _ = torch.max(attention.view(x.size(0), -1), dim=1, keepdim=True)
        threshold = max_val * np.random.uniform(0.7, 0.9)
        threshold = threshold.view(x.size(0), 1, 1, 1).expand_as(attention)
        drop_mask = (attention < threshold).float()
        return x.mul(drop_mask)

    def feature_based_noise(self, x):
        noise_vector = self.uni_dist.sample(x.shape[1:]).to(x.device).unsqueeze(0)
        x_noise = x.mul(noise_vector) + x
        return x_noise

    def forward(self, x):
        if(self.aux_type == "DropOut"):
            pass
        elif(self.aux_type == "FeatureDrop"):
            x[-1] = self.feature_drop(x[-1])
        elif(self.aux_type == "FeatureNoise"):
            x[-1] = self.feature_based_noise(x[-1])
        elif(self.aux_type == "VAT"):
            it = self.params.get("VAT_it".lower(), 2)
            xi = self.params.get("VAT_xi".lower(), 1e-6)
            eps= self.params.get("VAT_eps".lower(), 2.0)
            x[-1] = get_r_adv(x, self.decoder, it, xi, eps)
        else:
            raise ValueError("Undefined auxiliary decoder type {0:}".format(self.aux_type))
 
        output = self.decoder(x)
        return output


class UNet2D_CCT(nn.Module):
    def __init__(self, params):
        super(UNet2D_CCT, self).__init__()
        self.params    = params
        self.encoder   = Encoder(params)
        self.decoder   = Decoder(params)
        aux_names = params.get("CCT_aux_decoders".lower(), None)
        if aux_names is None:
            aux_names = ["DropOut", "FeatureDrop", "FeatureNoise", "VAT"]
        aux_decoders = []
        for aux_name in aux_names:
            aux_decoders.append(AuxiliaryDecoder(params, aux_name))
        self.aux_decoders = nn.ModuleList(aux_decoders)
        

    def forward(self, x):
        x_shape = list(x.shape)
        if(len(x_shape) == 5):
          [N, C, D, H, W] = x_shape
          new_shape = [N*D, C, H, W]
          x = torch.transpose(x, 1, 2)
          x = torch.reshape(x, new_shape)

        f = self.encoder(x)
        output = self.decoder(f)
        if(len(x_shape) == 5):
            new_shape = [N, D] + list(output.shape)[1:]
            output = torch.reshape(output, new_shape)
            output = torch.transpose(output, 1, 2)

        if(self.training):
            aux_outputs = [aux_d(f) for aux_d in self.aux_decoders]
            if(len(x_shape) == 5):
                for i in range(len(aux_outputs)):
                    aux_outi = torch.reshape(aux_outputs[i], new_shape)
                    aux_outputs[i] = torch.transpose(aux_outi, 1, 2)
            return output, aux_outputs
        else:
            return output


# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

import os
import csv 
import sys
import math
import pandas as pd
import random
import configparser
import numpy as np
import sklearn.metrics as metrics
from scipy import ndimage
from pymic.io.image_read_write import *
from pymic.util.image_process import *
from pymic.util.parse_config import parse_config

def accuracy(gt_label, pred_label):
    correct_pred = gt_label == pred_label
    acc = (correct_pred.sum() + 0.0 ) / len(gt_label)
    return acc 

def sensitivity(gt_label, pred_label):
    pos_pred = gt_label * pred_label
    senst = (pos_pred.sum() + 0.0) / gt_label.sum() 
    return senst

def specificity(gt_label, pred_label):
    gt_label = 1 - gt_label
    pred_label = 1 - pred_label
    neg_pred = gt_label * pred_label
    spec = (neg_pred.sum() + 0.0) / gt_label.sum() 
    return spec 

def get_evaluation_score(gt_label, pred_prob, metric):
    """
    the gt_label is 1-d array
    currently only binary classification is considered
    """
    pred_lab = np.argmax(pred_prob, axis = 1)
    if(metric == "accuracy"):
        score = metrics.accuracy_score(gt_label, pred_lab)
    elif(metric == "recall" or metric == "sensitivity"):
        score = metrics.recall_score(gt_label, pred_lab)
    elif(metric == "specificity"):
        score = metrics.recall_score(1 - gt_label, 1 - pred_lab)
    elif(metric == "precision"):
        score = metrics.precision_score(gt_label, pred_lab)
    elif(metric == "auc"):
        score = metrics.roc_auc_score(gt_label, pred_prob[:,1])
    else:
        raise ValueError("undefined metric: {0:}".format(metric))
    return score

def binary_evaluation(config):
    metric_list = config['metric_list']
    gt_csv  = config['ground_truth_csv']
    prob_csv= config['predict_prob_csv']
    gt_items  = pd.read_csv(gt_csv)
    prob_items = pd.read_csv(prob_csv)
    assert(len(gt_items) == len(prob_items))
    for i in range(len(gt_items)):
        assert(gt_items.iloc[i, 0] == prob_items.iloc[i, 0])
    
    gt_data  = np.asarray(gt_items.iloc[:, 1])
    prob_data = np.asarray(prob_items.iloc[:, 1:])
    score_list = []
    for metric in metric_list:
        score = get_evaluation_score(gt_data, prob_data, metric)
        score_list.append(score)
        print("{0:}: {1:}".format(metric, score))

    out_csv = prob_csv.replace("prob", "eval")
    with open(out_csv, mode='w') as csv_file:
        csv_writer = csv.writer(csv_file, delimiter=',', 
                            quotechar='"',quoting=csv.QUOTE_MINIMAL)
        csv_writer.writerow(metric_list)
        csv_writer.writerow(score_list)

def nexcl_evaluation(config):
    """
    evaluation for nonexclusive classification
    """
    metric_list = config['metric_list']
    gt_csv    = config['ground_truth_csv']
    prob_csv  = config['predict_prob_csv']
    gt_items  = pd.read_csv(gt_csv)
    prob_items= pd.read_csv(prob_csv)
    assert(len(gt_items) == len(prob_items))
    for i in range(len(gt_items)):
        assert(gt_items.iloc[i, 0] == prob_items.iloc[i, 0])
    
    cls_names = gt_items.columns[1:]
    cls_num   = len(cls_names)
    gt_data  = np.asarray(gt_items.iloc[:, 1:cls_num + 1])
    prob_data = np.asarray(prob_items.iloc[:, 1:cls_num + 1])
    score_list= []
    for metric in metric_list:
        print(metric)
        score_m = []
        for c in range(cls_num):
            gt_data_c = gt_data[:, c:c+1]
            prob_c = prob_data[:, c]
            prob_c = np.asarray([1.0 - prob_c, prob_c])
            prob_c = np.transpose(prob_c)
            score = get_evaluation_score(gt_data_c, prob_c, metric)
            score_m.append(score)
            print(cls_names[c], score)
        score_avg = np.asarray(score_m).mean()
        print('avg', score_avg)
        score_m.append(score_avg)
        score_list.append(score_m)

    out_csv = prob_csv.replace("prob", "eval")
    with open(out_csv, mode='w') as csv_file:
        csv_writer = csv.writer(csv_file, delimiter=',', 
                            quotechar='"',quoting=csv.QUOTE_MINIMAL)
        csv_writer.writerow(['metric'] + list(cls_names) + ['avg'])
        for i in range(len(score_list)):
            item = metric_list[i : i+1] + score_list[i]
            csv_writer.writerow(item)

def main():
    if(len(sys.argv) < 2):
        print('Number of arguments should be 2. e.g.')
        print('    pymic_evaluate_cls config.cfg')
        exit()
    config_file = str(sys.argv[1])
    assert(os.path.isfile(config_file))
    config = parse_config(config_file)['evaluation']
    task_type = config.get('task_type', "cls")
    if(task_type == "cls"):  # default exclusive classification
        binary_evaluation(config)
    else:                    # non exclusive classification
        nexcl_evaluation(config)
    
if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from __future__ import print_function, division
import logging
from pymic.net_run.agent_seg import SegmentationAgent

class WSLSegAgent(SegmentationAgent):
    """
    Training and testing agent for weakly supervised segmentation
    """
    def __init__(self, config, stage = 'train'):
        super(WSLSegAgent, self).__init__(config, stage)

    def training(self):
        pass
        
    def write_scalars(self, train_scalars, valid_scalars, lr_value, glob_it):
        loss_scalar ={'train':train_scalars['loss'], 
                      'valid':valid_scalars['loss']}
        loss_sup_scalar  = {'train':train_scalars['loss_sup']}
        loss_upsup_scalar  = {'train':train_scalars['loss_reg']}
        dice_scalar ={'train':train_scalars['avg_dice'], 'valid':valid_scalars['avg_dice']}
        self.summ_writer.add_scalars('loss', loss_scalar, glob_it)
        self.summ_writer.add_scalars('loss_sup', loss_sup_scalar, glob_it)
        self.summ_writer.add_scalars('loss_reg', loss_upsup_scalar, glob_it)
        self.summ_writer.add_scalars('regular_w', {'regular_w':train_scalars['regular_w']}, glob_it)
        self.summ_writer.add_scalars('lr', {"lr": lr_value}, glob_it)
        self.summ_writer.add_scalars('dice', dice_scalar, glob_it)
        class_num = self.config['network']['class_num']
        for c in range(class_num):
            cls_dice_scalar = {'train':train_scalars['class_dice'][c], \
                'valid':valid_scalars['class_dice'][c]}
            self.summ_writer.add_scalars('class_{0:}_dice'.format(c), cls_dice_scalar, glob_it)
        logging.info('train loss {0:.4f}, avg dice {1:.4f} '.format(
            train_scalars['loss'], train_scalars['avg_dice']) + "[" + \
            ' '.join("{0:.4f}".format(x) for x in train_scalars['class_dice']) + "]")        
        logging.info('valid loss {0:.4f}, avg dice {1:.4f} '.format(
            valid_scalars['loss'], valid_scalars['avg_dice']) + "[" + \
            ' '.join("{0:.4f}".format(x) for x in valid_scalars['class_dice']) + "]")  



# -*- coding: utf-8 -*-
"""
the original implementation is from:
https://github.com/LEONOB2014/GatedCRFLoss/blob/master/models/model_loss_semseg_gatedcrf.py
"""
import torch
import torch.nn.functional as F


class ModelLossSemsegGatedCRF(torch.nn.Module):
    """
    This module provides an implementation of the Gated CRF Loss for Weakly Supervised Semantic Image Segmentation.
    This loss function promotes consistent label assignment guided by input features, such as RGBXY.
    Please consider using the following bibtex for citation:
    @article{obukhov2019gated,
        author={Anton Obukhov and Stamatios Georgoulis and Dengxin Dai and Luc {Van Gool}},
        title={Gated {CRF} Loss for Weakly Supervised Semantic Image Segmentation},
        journal={CoRR},
        volume={abs/1906.04651},
        year={2019},
        url={http://arxiv.org/abs/1906.04651},
    }
    """
    def forward(
            self, y_hat_softmax, kernels_desc, kernels_radius, sample, height_input, width_input,
            mask_src=None, mask_dst=None, compatibility=None, custom_modality_downsamplers=None, out_kernels_vis=False
    ):
        """
        Performs the forward pass of the loss.
        :param y_hat_softmax: A tensor of predicted per-pixel class probabilities of size NxCxHxW
        :param kernels_desc: A list of dictionaries, each describing one Gaussian kernel composition from modalities.
            The final kernel is a weighted sum of individual kernels. Following example is a composition of
            RGBXY and XY kernels:
            kernels_desc: [{
                'weight': 0.9,          # Weight of RGBXY kernel
                'xy': 6,                # Sigma for XY
                'rgb': 0.1,             # Sigma for RGB
            },{
                'weight': 0.1,          # Weight of XY kernel
                'xy': 6,                # Sigma for XY
            }]
        :param kernels_radius: Defines size of bounding box region around each pixel in which the kernel is constructed.
        :param sample: A dictionary with modalities (except 'xy') used in kernels_desc parameter. Each of the provided
            modalities is allowed to be larger than the shape of y_hat_softmax, in such case downsampling will be
            invoked. Default downsampling method is area resize; this can be overriden by setting.
            custom_modality_downsamplers parameter.
        :param width_input, height_input: Dimensions of the full scale resolution of modalities
        :param mask_src: (optional) Source mask.
        :param mask_dst: (optional) Destination mask.
        :param compatibility: (optional) Classes compatibility matrix, defaults to Potts model.
        :param custom_modality_downsamplers: A dictionary of modality downsampling functions.
        :param out_kernels_vis: Whether to return a tensor with kernels visualized with some step.
        :return: Loss function value.
        """
        assert y_hat_softmax.dim() == 4, 'Prediction must be a NCHW batch'
        N, C, height_pred, width_pred = y_hat_softmax.shape
        device = y_hat_softmax.device

        assert width_input % width_pred == 0 and height_input % height_pred == 0 and \
               width_input * height_pred == height_input * width_pred, \
            f'[{width_input}x{height_input}] !~= [{width_pred}x{height_pred}]'

        kernels = self._create_kernels(
            kernels_desc, kernels_radius, sample, N, height_pred, width_pred, device, custom_modality_downsamplers
        )

        denom = N * height_pred * width_pred

        def resize_fix_mask(mask, name):
            assert mask.dim() == 4 and mask.shape[:2] == (N, 1) and mask.dtype == torch.float32, \
                f'{name} mask must be a NCHW batch with C=1 and dtype float32'
            if mask.shape[2:] != (height_pred, width_pred):
                mask = ModelLossSemsegGatedCRF._downsample(
                    mask, 'mask', height_pred, width_pred, custom_modality_downsamplers
                )
            mask[mask != mask] = 0.0    # handle NaN
            mask[mask < 1.0] = 0.0      # handle edges of mask after interpolation
            return mask

        if mask_src is not None:
            mask_src = resize_fix_mask(mask_src, 'Source')
            denom = mask_src.sum().clamp(min=1)
            mask_src = self._unfold(mask_src, kernels_radius)
            kernels = kernels * mask_src

        if mask_dst is not None:
            mask_dst = resize_fix_mask(mask_dst, 'Destination')
            denom = mask_dst.sum().clamp(min=1)
            mask_dst = mask_dst.view(N, 1, 1, 1, height_pred, width_pred)
            kernels = kernels * mask_dst

        y_hat_unfolded = self._unfold(y_hat_softmax, kernels_radius)

        product_kernel_x_y_hat = (kernels * y_hat_unfolded) \
            .view(N, C, (kernels_radius * 2 + 1) ** 2, height_pred, width_pred) \
            .sum(dim=2, keepdim=False)

        if compatibility is None:
            # Using shortcut for Pott's class compatibility model
            loss = -(product_kernel_x_y_hat * y_hat_softmax).sum()
            loss = kernels.sum() + loss  # comment out to save computation, total loss may go below 0
        else:
            assert compatibility.shape == (C, C), f'Compatibility matrix expected shape [{C}x{C}]'
            assert (compatibility < 0).int().sum() == 0, 'Compatibility matrix must not have negative values'
            assert compatibility.diag.sum() == 0, 'Compatibility matrix diagonal must be 0'
            compat = (C-1) * F.normalize(compatibility.float().to(device), p=1, dim=1)
            y_hat_CxNHW = y_hat_softmax.permute(1, 0, 2, 3).contiguous().view(C, -1)
            product_kernel_x_y_hat_NHWxC = product_kernel_x_y_hat.permute(0, 2, 3, 1).contiguous().view(-1, C)
            product_CxC = torch.mm(y_hat_CxNHW, product_kernel_x_y_hat_NHWxC)
            loss = (compat * product_CxC).sum()

        out = {
            'loss': loss / denom,
        }

        if out_kernels_vis:
            out['kernels_vis'] = self._visualize_kernels(
                kernels, kernels_radius, height_input, width_input, height_pred, width_pred
            )

        return out

    @staticmethod
    def _downsample(img, modality, height_dst, width_dst, custom_modality_downsamplers):
        if custom_modality_downsamplers is not None and modality in custom_modality_downsamplers:
            f_down = custom_modality_downsamplers[modality]
        else:
            f_down = F.adaptive_avg_pool2d
        return f_down(img, (height_dst, width_dst))

    @staticmethod
    def _create_kernels(
            kernels_desc, kernels_radius, sample, N, height_pred, width_pred, device, custom_modality_downsamplers
    ):
        kernels = None
        for i, desc in enumerate(kernels_desc):
            weight = desc['weight']
            features = []
            for modality, sigma in desc.items():
                if modality == 'weight':
                    continue
                if modality == 'xy':
                    feature = ModelLossSemsegGatedCRF._get_mesh(N, height_pred, width_pred, device)
                else:
                    assert modality in sample, \
                        f'Modality {modality} is listed in {i}-th kernel descriptor, but not present in the sample'
                    feature = sample[modality]
                    feature = ModelLossSemsegGatedCRF._downsample(
                        feature, modality, height_pred, width_pred, custom_modality_downsamplers
                    )
                feature /= sigma
                features.append(feature)
            features = torch.cat(features, dim=1)
            kernel = weight * ModelLossSemsegGatedCRF._create_kernels_from_features(features, kernels_radius)
            kernels = kernel if kernels is None else kernel + kernels
        return kernels

    @staticmethod
    def _create_kernels_from_features(features, radius):
        assert features.dim() == 4, 'Features must be a NCHW batch'
        N, C, H, W = features.shape
        kernels = ModelLossSemsegGatedCRF._unfold(features, radius)
        kernels = kernels - kernels[:, :, radius, radius, :, :].view(N, C, 1, 1, H, W)
        kernels = (-0.5 * kernels ** 2).sum(dim=1, keepdim=True).exp()
        kernels[:, :, radius, radius, :, :] = 0
        return kernels

    @staticmethod
    def _get_mesh(N, H, W, device):
        return torch.cat((
            torch.arange(0, W, 1, dtype=torch.float32, device=device).view(1, 1, 1, W).repeat(N, 1, H, 1),
            torch.arange(0, H, 1, dtype=torch.float32, device=device).view(1, 1, H, 1).repeat(N, 1, 1, W)
        ), 1)

    @staticmethod
    def _unfold(img, radius):
        assert img.dim() == 4, 'Unfolding requires NCHW batch'
        N, C, H, W = img.shape
        diameter = 2 * radius + 1
        return F.unfold(img, diameter, 1, radius).view(N, C, diameter, diameter, H, W)

    @staticmethod
    def _visualize_kernels(kernels, radius, height_input, width_input, height_pred, width_pred):
        diameter = 2 * radius + 1
        vis = kernels[:, :, :, :, radius::diameter, radius::diameter]
        vis_nh, vis_nw = vis.shape[-2:]
        vis = vis.permute(0, 1, 4, 2, 5, 3).contiguous().view(kernels.shape[0], 1, diameter * vis_nh, diameter * vis_nw)
        if vis.shape[2] > height_pred:
            vis = vis[:, :, :height_pred, :]
        if vis.shape[3] > width_pred:
            vis = vis[:, :, :, :width_pred]
        if vis.shape[2:] != (height_pred, width_pred):
            vis = F.pad(vis, [0, width_pred-vis.shape[3], 0, height_pred-vis.shape[2]])
        vis = F.interpolate(vis, (height_input, width_input), mode='nearest')
        return vis



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
from torch.nn.functional import interpolate

class Inferer(object):
    def __init__(self, config):
        self.config = config
        
    def __infer(self, image):
        use_sw  = self.config.get('sliding_window_enable', False)
        if(not use_sw):
            outputs = self.model(image)
        else:
            outputs = self.__infer_with_sliding_window(image)
        return outputs

    def __get_prediction_number_and_scales(self, tempx):
        """
        If the network outputs multiple tensors with different sizes, return the
        number of tensors and the scale of each tensor compared with the first one
        """
        img_dim = len(tempx.shape) - 2
        output = self.model(tempx)
        if(isinstance(output, (tuple, list))):
            output_num = len(output)
            scales = [[1.0] * img_dim]
            shape0 = list(output[0].shape[2:])
            for  i in range(1, output_num):
                shapei= list(output[i].shape[2:])
                scale = [(shapei[d] + 0.0) / shape0[d] for d in range(img_dim)]
                scales.append(scale)
        else:
            output_num, scales = 1, None
        return output_num, scales

    def __infer_with_sliding_window(self, image):
        """
        Use sliding window to predict segmentation for large images.
        Note that the network may output a list of tensors with difference sizes.
        """
        window_size   = [x for x in self.config['sliding_window_size']]
        window_stride = [x for x in self.config['sliding_window_stride']]
        class_num     = self.config['class_num']
        img_full_shape = list(image.shape)
        batch_size = img_full_shape[0]
        img_shape  = img_full_shape[2:]
        img_dim    = len(img_shape)
        if(img_dim != 2 and img_dim !=3):
            raise ValueError("Inference using sliding window only supports 2D and 3D images")

        for d in range(img_dim):
            if (window_size[d] is None) or window_size[d] > img_shape[d]:
                window_size[d]  = img_shape[d]
            if (window_stride[d] is None) or window_stride[d] > window_size[d]:
                window_stride[d] = window_size[d]
                
        if all([window_size[d] >= img_shape[d] for d in range(img_dim)]):
            output = self.model(image)
            return output

        crop_start_list  = []
        for w in range(0, img_shape[-1], window_stride[-1]):
            w_min = min(w, img_shape[-1] - window_size[-1])
            for h in range(0, img_shape[-2], window_stride[-2]):
                h_min = min(h, img_shape[-2] - window_size[-2])
                if(img_dim == 2):
                    crop_start_list.append([h_min, w_min])
                else:
                    for d in range(0, img_shape[0], window_stride[0]):
                        d_min = min(d, img_shape[0] - window_size[0])
                        crop_start_list.append([d_min, h_min, w_min])

        output_shape = [batch_size, class_num] + img_shape
        mask_shape   = [batch_size, class_num] + window_size
        counter      = torch.zeros(output_shape).to(image.device)
        temp_mask    = torch.ones(mask_shape).to(image.device)
        temp_in_shape = img_full_shape[:2] + window_size
        tempx = torch.ones(temp_in_shape).to(image.device)
        out_num, scale_list = self.__get_prediction_number_and_scales(tempx)
        if(out_num == 1): # for a single prediction
            output = torch.zeros(output_shape).to(image.device)
            for c0 in crop_start_list:
                c1 = [c0[d] + window_size[d] for d in range(img_dim)]
                if(img_dim == 2):
                    patch_in = image[:, :, c0[0]:c1[0], c0[1]:c1[1]]
                else:
                    patch_in = image[:, :, c0[0]:c1[0], c0[1]:c1[1], c0[2]:c1[2]]
                patch_out = self.model(patch_in) 
                if(isinstance(patch_out, (tuple, list))):
                    patch_out = patch_out[0]
                if(img_dim == 2):
                    output[:, :, c0[0]:c1[0], c0[1]:c1[1]] += patch_out
                    counter[:, :, c0[0]:c1[0], c0[1]:c1[1]] += temp_mask
                else:
                    output[:, :, c0[0]:c1[0], c0[1]:c1[1], c0[2]:c1[2]] += patch_out
                    counter[:, :, c0[0]:c1[0], c0[1]:c1[1], c0[2]:c1[2]] += temp_mask
            return output/counter
        else: # for multiple prediction
            output_list= []
            for i in range(out_num):
                output_shape_i = [batch_size, class_num] + \
                    [int(img_shape[d] * scale_list[i][d]) for d in range(img_dim)]
                output_list.append(torch.zeros(output_shape_i).to(image.device))

            for c0 in crop_start_list:
                c1 = [c0[d] + window_size[d] for d in range(img_dim)]
                if(img_dim == 2):
                    patch_in = image[:, :, c0[0]:c1[0], c0[1]:c1[1]]
                else:
                    patch_in = image[:, :, c0[0]:c1[0], c0[1]:c1[1], c0[2]:c1[2]]
                patch_out = self.model(patch_in) 

                for i in range(out_num):
                    c0_i = [int(c0[d] * scale_list[i][d]) for d in range(img_dim)]
                    c1_i = [int(c1[d] * scale_list[i][d]) for d in range(img_dim)]
                    if(img_dim == 2):
                        output_list[i][:, :, c0_i[0]:c1_i[0], c0_i[1]:c1_i[1]] += patch_out[i]
                        counter[:, :, c0[0]:c1[0], c0[1]:c1[1]] += temp_mask
                    else:
                        output_list[i][:, :, c0_i[0]:c1_i[0], c0_i[1]:c1_i[1], c0_i[2]:c1_i[2]] += patch_out[i]
                        counter[:, :, c0[0]:c1[0], c0[1]:c1[1], c0[2]:c1[2]] += temp_mask
            for i in range(out_num):  
                counter_i = interpolate(counter, scale_factor = scale_list[i])
                output_list[i] = output_list[i] / counter_i
            return output_list

    def run(self, model, image):
        self.model = model
        tta_mode   = self.config.get('tta_mode', 0)
        if(tta_mode == 0):
            outputs = self.__infer(image)
        elif(tta_mode == 1): 
            # test time augmentation with flip in 2D
            # you may define your own method for test time augmentation
            outputs1 = self.__infer(image)
            outputs2 = self.__infer(torch.flip(image, [-2]))
            outputs3 = self.__infer(torch.flip(image, [-1]))
            outputs4 = self.__infer(torch.flip(image, [-2, -1]))
            if(isinstance(outputs1, (tuple, list))):
                outputs = []
                for i in range(len(outputs1)):
                    temp_out1 = outputs1[i]
                    temp_out2 = torch.flip(outputs2[i], [-2])
                    temp_out3 = torch.flip(outputs3[i], [-1])
                    temp_out4 = torch.flip(outputs4[i], [-2, -1])
                    temp_mean = (temp_out1 + temp_out2 + temp_out3 + temp_out4) / 4
                    outputs.append(temp_mean)
            else:
                outputs2 = torch.flip(outputs2, [-2])
                outputs3 = torch.flip(outputs3, [-1])
                outputs4 = torch.flip(outputs4, [-2, -1])
                outputs = (outputs1 + outputs2 + outputs3 + outputs4) / 4
        else:
            raise ValueError("Undefined tta_mode {0:}".format(tta_mode))
        return outputs




# -*- coding: utf-8 -*-
from __future__ import print_function, division
import logging
import torch
import numpy as np
from torch.optim import lr_scheduler
from pymic.loss.seg.util import get_soft_label
from pymic.loss.seg.util import reshape_prediction_and_ground_truth
from pymic.loss.seg.util import get_classwise_dice
from pymic.net_run_ssl.ssl_abstract import SSLSegAgent
from pymic.net.net_dict_seg import SegNetDict
from pymic.util.ramps import get_rampup_ratio

class SSLMeanTeacher(SSLSegAgent):
    """
    Mean Teacher for semi-supervised learning according to the following paper:
    Antti Tarvainen, Harri Valpola: Mean teachers are better role models: Weight-averaged 
    consistency targets improve semi-supervised deep learning results.
    NeurIPS 2017.
    https://arxiv.org/abs/1703.01780
    """
    def __init__(self, config, stage = 'train'):
        super(SSLMeanTeacher, self).__init__(config, stage)
        self.net_ema = None 

    def create_network(self):
        super(SSLMeanTeacher, self).create_network()
        if(self.net_ema is None):
            net_name = self.config['network']['net_type']
            if(net_name not in SegNetDict):
                raise ValueError("Undefined network {0:}".format(net_name))
            self.net_ema = SegNetDict[net_name](self.config['network'])
        if(self.tensor_type == 'float'):
            self.net_ema.float()
        else:
            self.net_ema.double()

    def training(self):
        class_num   = self.config['network']['class_num']
        iter_valid  = self.config['training']['iter_valid']
        ssl_cfg     = self.config['semi_supervised_learning']
        iter_max     = self.config['training']['iter_max']
        rampup_start = ssl_cfg.get('rampup_start', 0)
        rampup_end   = ssl_cfg.get('rampup_end', iter_max)
        train_loss  = 0
        train_loss_sup = 0
        train_loss_reg = 0
        train_dice_list = []
        self.net.train()
        self.net_ema.to(self.device)
        for it in range(iter_valid):
            try:
                data_lab = next(self.trainIter)
            except StopIteration:
                self.trainIter = iter(self.train_loader)
                data_lab = next(self.trainIter)
            try:
                data_unlab = next(self.trainIter_unlab)
            except StopIteration:
                self.trainIter_unlab = iter(self.train_loader_unlab)
                data_unlab = next(self.trainIter_unlab)

            # get the inputs
            x0   = self.convert_tensor_type(data_lab['image'])
            y0   = self.convert_tensor_type(data_lab['label_prob'])  
            x1   = self.convert_tensor_type(data_unlab['image'])
            inputs = torch.cat([x0, x1], dim = 0)               
            inputs, y0 = inputs.to(self.device), y0.to(self.device)
            noise = torch.clamp(torch.randn_like(x1) * 0.1, -0.2, 0.2)
            inputs_ema = x1 + torch.clamp(torch.randn_like(x1) * 0.1, -0.2, 0.2)
            inputs_ema = inputs_ema.to(self.device)

            # zero the parameter gradients
            self.optimizer.zero_grad()
                
            outputs = self.net(inputs)
            n0 = list(x0.shape)[0] 
            p0 = outputs[:n0]
            loss_sup = self.get_loss_value(data_lab, p0, y0)

            outputs_soft = torch.softmax(outputs, dim=1)
            p1_soft = outputs_soft[n0:]
            
            with torch.no_grad():
                outputs_ema = self.net_ema(inputs_ema)
                p1_ema_soft = torch.softmax(outputs_ema, dim=1)
            
            rampup_ratio = get_rampup_ratio(self.glob_it, rampup_start, rampup_end, "sigmoid")
            regular_w = ssl_cfg.get('regularize_w', 0.1) * rampup_ratio

            loss_reg = torch.nn.MSELoss()(p1_soft, p1_ema_soft)
            loss = loss_sup + regular_w*loss_reg

            loss.backward()
            self.optimizer.step()
            if(self.scheduler is not None and \
                not isinstance(self.scheduler, lr_scheduler.ReduceLROnPlateau)):
                self.scheduler.step()

            # update EMA
            alpha = ssl_cfg.get('ema_decay', 0.99)
            alpha = min(1 - 1 / (iter_max + 1), alpha)
            for ema_param, param in zip(self.net_ema.parameters(), self.net.parameters()):
                ema_param.data.mul_(alpha).add_(1 - alpha, param.data)

            train_loss = train_loss + loss.item()
            train_loss_sup = train_loss_sup + loss_sup.item()
            train_loss_reg = train_loss_reg + loss_reg.item() 
            # get dice evaluation for each class in annotated images
            if(isinstance(p0, tuple) or isinstance(p0, list)):
                p0 = p0[0] 
            p0_argmax = torch.argmax(p0, dim = 1, keepdim = True)
            p0_soft   = get_soft_label(p0_argmax, class_num, self.tensor_type)
            p0_soft, y0 = reshape_prediction_and_ground_truth(p0_soft, y0) 
            dice_list   = get_classwise_dice(p0_soft, y0)
            train_dice_list.append(dice_list.cpu().numpy())
        train_avg_loss = train_loss / iter_valid
        train_avg_loss_sup = train_loss_sup / iter_valid
        train_avg_loss_reg = train_loss_reg / iter_valid
        train_cls_dice = np.asarray(train_dice_list).mean(axis = 0)
        train_avg_dice = train_cls_dice.mean()

        train_scalers = {'loss': train_avg_loss, 'loss_sup':train_avg_loss_sup,
            'loss_reg':train_avg_loss_reg, 'regular_w':regular_w,
            'avg_dice':train_avg_dice,     'class_dice': train_cls_dice}
        return train_scalers


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn
from pymic.net.net2d.unet2d import *
"""
A Reimplementation of the attention U-Net paper:
    Ozan Oktay, Jo Schlemper et al.:
    Attentin U-Net: Looking Where to Look for the Pancreas. MIDL, 2018.

Note that there are some modifications from the original paper, such as
the use of batch normalization, dropout, and leaky relu here.
"""
class AttentionGateBlock(nn.Module):
    def __init__(self, chns_l, chns_h):
        """
        chns_l: channel number of low-level features from the encoder
        chns_h: channel number of high-level features from the decoder
        """
        super(AttentionGateBlock, self).__init__()
        self.in_chns_l = chns_l
        self.in_chns_h = chns_h

        self.out_chns = int(min(self.in_chns_l, self.in_chns_h)/2)
        self.conv1_l = nn.Conv2d(self.in_chns_l, self.out_chns,
                kernel_size = 1, bias = True)
        self.conv1_h = nn.Conv2d(self.in_chns_h, self.out_chns,
                kernel_size = 1, bias = True)
        self.conv2 = nn.Conv2d(self.out_chns, 1,
                kernel_size = 1, bias = True)
        self.act1 = nn.ReLU()
        self.act2 = nn.Sigmoid()

    def forward(self, x_l, x_h):
        input_shape = list(x_l.shape)
        gate_shape  = list(x_h.shape)

        # resize low-level feature to the shape of high-level feature
        x_l_reshape = nn.functional.interpolate(x_l, size = gate_shape[2:], mode = 'bilinear')
        f_l = self.conv1_l(x_l_reshape)
        f_h = self.conv1_h(x_h)
        f = f_l + f_h
        f = self.act1(f)
        f = self.conv2(f)
        att = self.act2(f)
        # resize attention map to the shape of low-level feature
        att = nn.functional.interpolate(att, size = input_shape[2:], mode = 'bilinear')
        # return calibrated low-level feature
        output = att * x_l
        return output


class UpBlockWithAttention(nn.Module):
    """Upsampling followed by ConvBlock"""
    def __init__(self, in_channels1, in_channels2, out_channels, dropout_p,
                 bilinear=True):
        """
        in_channels1: channel of high-level features
        in_channels2: channel of low-level features
        out_channels: output channel number
        dropout_p: probability of dropout
        """
        super(UpBlockWithAttention, self).__init__()
        self.bilinear = bilinear
        if bilinear:
            self.conv1x1 = nn.Conv2d(in_channels1, in_channels2, kernel_size = 1)
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
        else:
            self.up = nn.ConvTranspose2d(in_channels1, in_channels2, kernel_size=2, stride=2)
        self.conv = ConvBlock(in_channels2 * 2, out_channels, dropout_p)
        self.ag   = AttentionGateBlock(in_channels2, in_channels1)

    def forward(self, x1, x2):
        x2at = self.ag(x2, x1)
        if self.bilinear:
            x1 = self.conv1x1(x1)
        x1 = self.up(x1)
        x = torch.cat([x2at, x1], dim=1)
        return self.conv(x)

class AttentionUNet2D(UNet2D):
    def __init__(self, params):
        super(AttentionUNet2D, self).__init__(params)
        self.up1    = UpBlockWithAttention(self.ft_chns[4], self.ft_chns[3], self.ft_chns[3], dropout_p = 0.0)
        self.up2    = UpBlockWithAttention(self.ft_chns[3], self.ft_chns[2], self.ft_chns[2], dropout_p = 0.0)
        self.up3    = UpBlockWithAttention(self.ft_chns[2], self.ft_chns[1], self.ft_chns[1], dropout_p = 0.0)
        self.up4    = UpBlockWithAttention(self.ft_chns[1], self.ft_chns[0], self.ft_chns[0], dropout_p = 0.0)





# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn
from pymic.loss.seg.util import reshape_tensor_to_2D, get_classwise_dice

class ExpLogLoss(nn.Module):
    """
    The exponential logarithmic loss in this paper: 
        K. Wong et al.: 3D Segmentation with Exponential Logarithmic Loss for Highly 
        Unbalanced Object Sizes. MICCAI 2018.
    """
    def __init__(self, params):
        super(ExpLogLoss, self).__init__()
        self.softmax = params.get('loss_softmax', True)
        self.w_dice = params['ExpLogLoss_w_dice'.lower()]
        self.gamma  = params['ExpLogLoss_gamma'.lower()]

    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        soft_y  = loss_input_dict['ground_truth']
        
        if(isinstance(predict, (list, tuple))):
            predict = predict[0]
        if(self.softmax):
            predict = nn.Softmax(dim = 1)(predict)
        predict = reshape_tensor_to_2D(predict)
        soft_y  = reshape_tensor_to_2D(soft_y)

        dice_score = get_classwise_dice(predict, soft_y)
        dice_score = 0.005 + dice_score * 0.99
        exp_dice   = -torch.log(dice_score)
        exp_dice   = torch.pow(exp_dice, self.gamma)
        exp_dice   = torch.mean(exp_dice)

        predict= 0.005 + predict * 0.99
        wc     = torch.mean(soft_y, dim = 0)
        wc     = 1.0 / (wc + 0.1)
        wc     = torch.pow(wc, 0.5)
        ce     = - torch.log(predict)
        exp_ce = wc * torch.pow(ce, self.gamma)
        exp_ce = torch.sum(soft_y * exp_ce, dim = 1)
        exp_ce = torch.mean(exp_ce)

        loss = exp_dice * self.w_dice + exp_ce * (1.0 - self.w_dice)
        return loss


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn
from pymic.loss.cls.util import get_soft_label

class L1Loss(nn.Module):
    """
    L1 (MAE) loss for classification
    """
    def __init__(self, params):
        super(L1Loss, self).__init__()
        self.l1_loss = nn.L1Loss()
    
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        labels  = loss_input_dict['ground_truth'][:, None] # reshape to N, 1
        softmax = nn.Softmax(dim = 1)
        predict = softmax(predict)
        num_class  = list(predict.size())[1]
        data_type = 'float' if(predict.dtype is torch.float32) else 'double'
        soft_y = get_soft_label(labels, num_class, data_type)
        loss = self.l1_loss(predict, soft_y)
        return loss

class RectifiedLoss(nn.Module):
    def __init__(self, params):
        super(RectifiedLoss, self).__init__()
        # self.l1_loss = nn.L1Loss()
    
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        labels  = loss_input_dict['ground_truth'][:, None] # reshape to N, 1
        
        # softmax = nn.Softmax(dim = 1)
        # predict = softmax(predict)
        num_class  = list(predict.size())[1]
        data_type = 'float' if(predict.dtype is torch.float32) else 'double'
        soft_y = get_soft_label(labels, num_class, data_type)
        g = 2* soft_y - 1
        loss = torch.exp((g*1.5- predict) * g)
        mask = predict < g 
        if (data_type  == 'float'):
            mask = mask.float() 
        else:
            mask = mask.double() 
        w = (mask - 0.5) * g + 0.5    
        loss = w * loss + 0.1*(g - predict) * (g - predict)
        loss = loss.mean()
        # loss = self.l1_loss(predict, soft_y)
        return loss



import torch

checkpoint_name1 = "/home/guotai/projects/PyMIC/examples/brats/model/casecade/wt/unet3d_4_8000.pt"
checkpoint1 = torch.load(checkpoint_name1)
state_dict1 = checkpoint1['model_state_dict']

checkpoint_name2 = "/home/guotai/projects/PyMIC/examples/brats/model/casecade/wt/unet3d_4_10000.pt"
checkpoint2 = torch.load(checkpoint_name2)
state_dict2 = checkpoint2['model_state_dict']

checkpoint_name3 = "/home/guotai/projects/PyMIC/examples/brats/model/casecade/wt/unet3d_4_12000.pt"
checkpoint3 = torch.load(checkpoint_name3)
state_dict3 = checkpoint3['model_state_dict']

state_dict = {}
for item in state_dict1:
    print(item)
    state_dict[item] = (state_dict1[item] + state_dict2[item] + state_dict3[item])/3

save_dict = {'model_state_dict': state_dict}
save_name = "/home/guotai/projects/PyMIC/examples/brats/model/casecade/wt/unet3d_4_avg.pt"
torch.save(save_dict, save_name) 


# -*- coding: utf-8 -*-
"""
Spatial Label Smoothing Regularization (SLSR) loss for learning from
noisy annotatins according to the following paper:
    Minqing Zhang, Jiantao Gao et al.:
    Characterizing Label Errors: Confident Learning for Noisy-Labeled Image 
    Segmentation, MICCAI 2020.
    https://link.springer.com/chapter/10.1007/978-3-030-59710-8_70 
"""
from __future__ import print_function, division

import torch
import torch.nn as nn
from pymic.loss.seg.util import reshape_tensor_to_2D

class SLSRLoss(nn.Module):
    def __init__(self, params):
        super(SLSRLoss, self).__init__()
        if(params is None):
            params = {}
        self.softmax = params.get('loss_softmax', True)
        self.epsilon = params.get('slsrloss_epsilon', 0.25)
    
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        soft_y  = loss_input_dict['ground_truth']
        pix_w   = loss_input_dict.get('pixel_weight', None)
        # the pixel wight here is actually the confidence mask
        # i.e., if the value is one, it means the label of corresponding 
        # pixel is noisy and should be replaced with smoothed label.

        if(isinstance(predict, (list, tuple))):
            predict = predict[0]
        if(self.softmax):
            predict = nn.Softmax(dim = 1)(predict)
        predict = reshape_tensor_to_2D(predict)
        soft_y  = reshape_tensor_to_2D(soft_y)
        if(pix_w is not None):
            pix_w   = reshape_tensor_to_2D(pix_w > 0).float()
            # smooth labels for pixels in the unconfident mask 
            smooth_y = (soft_y - 0.5) * (0.5 - self.epsilon) / 0.5 + 0.5
            smooth_y = pix_w * smooth_y + (1 - pix_w) * soft_y
        else:
            smooth_y = soft_y

        # for numeric stability
        predict = predict * 0.999 + 5e-4
        ce = - smooth_y* torch.log(predict)
        ce = torch.sum(ce, dim = 1) # shape is [N]
        ce = torch.mean(ce)  
        return ce



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import json
import math
import random
import numpy as np
from scipy import ndimage
from pymic.transform.abstract_transform import AbstractTransform
from pymic.util.image_process import *


class RandomFlip(AbstractTransform):
    """ random flip the image (shape [C, D, H, W] or [C, H, W]) """
    def __init__(self, params):
        """
            flip_depth (bool) : random flip along depth axis or not, only used for 3D images
            flip_height (bool): random flip along height axis or not
            flip_width (bool) : random flip along width axis or not
        """
        super(RandomFlip, self).__init__(params)
        self.flip_depth  = params['RandomFlip_flip_depth'.lower()]
        self.flip_height = params['RandomFlip_flip_height'.lower()]
        self.flip_width  = params['RandomFlip_flip_width'.lower()]
        self.inverse = params.get('RandomFlip_inverse'.lower(), True)

    def __call__(self, sample):
        image = sample['image']
        input_shape = image.shape
        input_dim = len(input_shape) - 1
        flip_axis = []
        if(self.flip_width):
            if(random.random() > 0.5):
                flip_axis.append(-1)
        if(self.flip_height):
            if(random.random() > 0.5):
                flip_axis.append(-2)
        if(input_dim == 3 and self.flip_depth):
            if(random.random() > 0.5):
                flip_axis.append(-3)

        sample['RandomFlip_Param'] = json.dumps(flip_axis)
        if(len(flip_axis) > 0):
            # use .copy() to avoid negative strides of numpy array
            # current pytorch does not support negative strides
            image_t = np.flip(image, flip_axis).copy()
            sample['image'] = image_t
            if('label' in sample and self.task == 'segmentation'):
                sample['label'] = np.flip(sample['label'] , flip_axis).copy()
            if('pixel_weight' in sample and self.task == 'segmentation'):
                sample['pixel_weight'] = np.flip(sample['pixel_weight'] , flip_axis).copy()
            
        return sample

    def  inverse_transform_for_prediction(self, sample):
        ''' flip sample['predict'] (5D or 4D) to the original direction.
        flip_axis is a list as saved in __call__().'''
        if(isinstance(sample['RandomFlip_Param'], list) or \
            isinstance(sample['RandomFlip_Param'], tuple)):
            flip_axis = json.loads(sample['RandomFlip_Param'][0]) 
        else:
            flip_axis = json.loads(sample['RandomFlip_Param']) 
        if(len(flip_axis) > 0):
            sample['predict']  = np.flip(sample['predict'] , flip_axis).copy()
        return sample


# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

import configparser
import logging

def is_int(val_str):
    start_digit = 0
    if(val_str[0] =='-'):
        start_digit = 1
    flag = True
    for i in range(start_digit, len(val_str)):
        if(str(val_str[i]) < '0' or str(val_str[i]) > '9'):
            flag = False
            break
    return flag

def is_float(val_str):
    flag = False
    if('.' in val_str and len(val_str.split('.'))==2 and not('./' in val_str)):
        if(is_int(val_str.split('.')[0]) and is_int(val_str.split('.')[1])):
            flag = True
        else:
            flag = False
    elif('e' in val_str and val_str[0] != 'e' and len(val_str.split('e'))==2):
        if(is_int(val_str.split('e')[0]) and is_int(val_str.split('e')[1])):
            flag = True
        else:
            flag = False       
    else:
        flag = False
    return flag 

def is_bool(var_str):
    if( var_str.lower() =='true' or var_str.lower() == 'false'):
        return True
    else:
        return False
    
def parse_bool(var_str):
    if(var_str.lower() =='true'):
        return True
    else:
        return False
     
def is_list(val_str):
    if(val_str[0] == '[' and val_str[-1] == ']'):
        return True
    else:
        return False
    
def parse_list(val_str):
    sub_str = val_str[1:-1]
    splits = sub_str.split(',')
    output = []
    for item in splits:
        item = item.strip()
        if(is_int(item)):
            output.append(int(item))
        elif(is_float(item)):
            output.append(float(item))
        elif(is_bool(item)):
            output.append(parse_bool(item))
        elif(item.lower() == 'none'):
            output.append(None)
        else:
            output.append(item)
    return output

def parse_value_from_string(val_str):
#     val_str = val_str.encode('ascii','ignore')
    if(is_int(val_str)):
        val = int(val_str)
    elif(is_float(val_str)):
        val = float(val_str)
    elif(is_list(val_str)):
        val = parse_list(val_str)
    elif(is_bool(val_str)):
        val = parse_bool(val_str)
    elif(val_str.lower() == 'none'):
        val = None
    else:
        val = val_str
    return val

def parse_config(filename):
    config = configparser.ConfigParser()
    config.read(filename)
    output = {}
    for section in config.sections():
        output[section] = {}
        for key in config[section]:
            val_str = str(config[section][key])
            if(len(val_str)>0):
                val = parse_value_from_string(val_str)
                output[section][key] = val
            else:
                val = None
            print(section, key, val)
    return output
            
def synchronize_config(config):
    data_cfg = config['dataset'] 
    net_cfg  = config['network']
    # data_cfg["modal_num"] = net_cfg["in_chns"]
    data_cfg["LabelToProbability_class_num".lower()] = net_cfg["class_num"] 
    if "PartialLabelToProbability" in data_cfg['train_transform']:
        data_cfg["PartialLabelToProbability_class_num".lower()] = net_cfg["class_num"]
    config['dataset'] = data_cfg
    config['network'] = net_cfg
    return config 

def logging_config(config):
    for section in config:
        for key in config[section]:
            value = config[section][key]
            logging.info("{0:} {1:} = {2:}".format(section, key, value))

if __name__ == "__main__":
    print(is_int('555'))
    print(is_float('555.10'))
    a='[1 ,2 ,3 ]'
    print(a)
    print(parse_list(a))
    
    



# -*- coding: utf-8 -*-
"""
An implementation of the Nested U-Net paper:
    Zongwei Zhou, et al.:
    UNet++: A Nested U-Net Architecture for Medical Image Segmentation. 
    MICCAI DLMIA workshop, 2018: 3-11.
Note that there are some modifications from the original paper, such as
the use of dropout and leaky relu here.
"""
import torch
import torch.nn as nn
from pymic.net.net2d.unet2d import *

class NestedUNet2D(nn.Module):
    def __init__(self, params):
        super(NestedUNet2D, self).__init__()
        self.params  = params
        self.in_chns = self.params['in_chns']
        self.filters = self.params['feature_chns'] 
        self.n_class = self.params['class_num']
        self.dropout = self.params['dropout']

        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.Up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)

        self.conv0_0 = ConvBlock(self.in_chns, self.filters[0], self.dropout[0])
        self.conv1_0 = ConvBlock(self.filters[0], self.filters[1], self.dropout[1])
        self.conv2_0 = ConvBlock(self.filters[1], self.filters[2], self.dropout[2])
        self.conv3_0 = ConvBlock(self.filters[2], self.filters[3], self.dropout[3])
        self.conv4_0 = ConvBlock(self.filters[3], self.filters[4], self.dropout[4])

        self.conv0_1 = ConvBlock(self.filters[0] + self.filters[1], self.filters[0], 0)
        self.conv1_1 = ConvBlock(self.filters[1] + self.filters[2], self.filters[1], 0)
        self.conv2_1 = ConvBlock(self.filters[2] + self.filters[3], self.filters[2], 0)
        self.conv3_1 = ConvBlock(self.filters[3] + self.filters[4], self.filters[3], 0)

        self.conv0_2 = ConvBlock(self.filters[0] * 2 + self.filters[1], self.filters[0], 0)
        self.conv1_2 = ConvBlock(self.filters[1] * 2 + self.filters[2], self.filters[1], 0)
        self.conv2_2 = ConvBlock(self.filters[2] * 2 + self.filters[3], self.filters[2], 0)

        self.conv0_3 = ConvBlock(self.filters[0] * 3 + self.filters[1], self.filters[0], 0)
        self.conv1_3 = ConvBlock(self.filters[1] * 3 + self.filters[2], self.filters[1], 0)

        self.conv0_4 = ConvBlock(self.filters[0] * 4 + self.filters[1], self.filters[0], 0)

        self.final = nn.Conv2d(self.filters[0], self.n_class, kernel_size=1)

    def forward(self, x):
        x_shape = list(x.shape)
        if(len(x_shape) == 5):
          [N, C, D, H, W] = x_shape
          new_shape = [N*D, C, H, W]
          x = torch.transpose(x, 1, 2)
          x = torch.reshape(x, new_shape)
        x0_0 = self.conv0_0(x)
        x1_0 = self.conv1_0(self.pool(x0_0))
        x0_1 = self.conv0_1(torch.cat([x0_0, self.Up(x1_0)], 1))

        x2_0 = self.conv2_0(self.pool(x1_0))
        x1_1 = self.conv1_1(torch.cat([x1_0, self.Up(x2_0)], 1))
        x0_2 = self.conv0_2(torch.cat([x0_0, x0_1, self.Up(x1_1)], 1))

        x3_0 = self.conv3_0(self.pool(x2_0))
        x2_1 = self.conv2_1(torch.cat([x2_0, self.Up(x3_0)], 1))
        x1_2 = self.conv1_2(torch.cat([x1_0, x1_1, self.Up(x2_1)], 1))
        x0_3 = self.conv0_3(torch.cat([x0_0, x0_1, x0_2, self.Up(x1_2)], 1))

        x4_0 = self.conv4_0(self.pool(x3_0))
        x3_1 = self.conv3_1(torch.cat([x3_0, self.Up(x4_0)], 1))
        x2_2 = self.conv2_2(torch.cat([x2_0, x2_1, self.Up(x3_1)], 1))
        x1_3 = self.conv1_3(torch.cat([x1_0, x1_1, x1_2, self.Up(x2_2)], 1))
        x0_4 = self.conv0_4(torch.cat([x0_0, x0_1, x0_2, x0_3, self.Up(x1_3)], 1))

        output = self.final(x0_4)
        if(len(x_shape) == 5):
            new_shape = [N, D] + list(output.shape)[1:]
            output = torch.reshape(output, new_shape)
            output = torch.transpose(output, 1, 2)
        return output


if __name__ == "__main__":
    params = {'in_chns':4,
              'feature_chns':[2, 8, 32, 48, 64],
              'dropout':  [0, 0, 0.3, 0.4, 0.5],
              'class_num': 2}
    Net = NestedUNet2D(params)
    Net = Net.double()

    x  = np.random.rand(4, 4, 10, 96, 96)
    xt = torch.from_numpy(x)
    xt = torch.tensor(xt)
    
    y = Net(xt)
    print(len(y.size()))
    y = y.detach().numpy()
    print(y.shape)



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import json
import math
import random
import numpy as np
from scipy import ndimage
from pymic.transform.abstract_transform import AbstractTransform
from pymic.util.image_process import *

class CenterCrop(AbstractTransform):
    """
    Crop the given image at the center.
    input shape should be [C, D, H, W] or [C, H, W])
    """
    def __init__(self, params):
        """
        output_size (tuple/list): Desired spatial output size. 
            [D, H, W] for 3D images and [H, W] for 2D images
            If D is None, then the z-axis is not cropped
        """
        self.output_size = params['CenterCrop_output_size'.lower()]
        self.inverse = params.get('CenterCrop_inverse'.lower(), True)
        self.task = params['Task'.lower()]

    def get_crop_param(self, sample):
        input_shape = sample['image'].shape
        input_dim   = len(input_shape) - 1
        assert(input_dim == len(self.output_size))
        temp_output_size = self.output_size
        if(input_dim == 3 and self.output_size[0] is None):
            # note that output size is [D, H, W] and input is [C, D, H, W]
            temp_output_size = [input_shape[1]] + self.output_size[1:]

        crop_margin = [input_shape[i + 1] - temp_output_size[i]\
            for i in range(input_dim)]
        crop_min = [int(item/2) for item in crop_margin]
        crop_max = [crop_min[i] + temp_output_size[i] \
            for i in range(input_dim)]
        crop_min = [0] + crop_min
        crop_max = list(input_shape[0:1]) + crop_max
        sample['CenterCrop_Param'] = json.dumps((input_shape, crop_min, crop_max))
        return sample, crop_min, crop_max

    def __call__(self, sample):
        image = sample['image']
        sample, crop_min, crop_max = self.get_crop_param(sample)

        image_t = crop_ND_volume_with_bounding_box(image, crop_min, crop_max)
        sample['image'] = image_t
        
        if('label' in sample and self.task == 'segmentation'):
            label = sample['label']
            crop_max[0] = label.shape[0]
            label = crop_ND_volume_with_bounding_box(label, crop_min, crop_max)
            sample['label'] = label
        if('pixel_weight' in sample and self.task == 'segmentation'):
            weight = sample['pixel_weight']
            crop_max[0] = weight.shape[0]
            weight = crop_ND_volume_with_bounding_box(weight, crop_min, crop_max)
            sample['pixel_weight'] = weight
        return sample

    def get_param_for_inverse_transform(self, sample):
        if(isinstance(sample['CenterCrop_Param'], list) or \
            isinstance(sample['CenterCrop_Param'], tuple)):
            params = json.loads(sample['CenterCrop_Param'][0]) 
        else:
            params = json.loads(sample['CenterCrop_Param']) 
        return params

    def inverse_transform_for_prediction(self, sample):
        ''' rescale sample['predict'] (5D or 4D) to the original spatial shape.
        origin_shape is a 4D or 3D vector as saved in __call__().'''
        params = self.get_param_for_inverse_transform(sample)
        origin_shape = params[0]
        crop_min     = params[1]
        crop_max     = params[2]
        predict = sample['predict']
        if(isinstance(predict, tuple) or isinstance(predict, list)):
            output_predict = []
            for predict_i in predict:
                origin_shape_i   = list(predict_i.shape[:2]) + origin_shape[1:]
                output_predict_i = np.zeros(origin_shape_i, predict_i.dtype)
                crop_min_i = [0, 0] + crop_min[1:]
                crop_max_i = list(predict_i.shape[:2]) + crop_max[1:]
                output_predict_i = set_ND_volume_roi_with_bounding_box_range(output_predict_i,
                    crop_min_i, crop_max_i, predict_i)
                output_predict.append(output_predict_i)
        else:
            origin_shape   = list(predict.shape[:2]) + origin_shape[1:]
            output_predict = np.zeros(origin_shape, predict.dtype)
            crop_min = [0, 0] + crop_min[1:]
            crop_max = list(predict.shape[:2]) + crop_max[1:]
            output_predict = set_ND_volume_roi_with_bounding_box_range(output_predict,
                crop_min, crop_max, predict)
        
        sample['predict'] = output_predict
        return sample

class CropWithBoundingBox(CenterCrop):
    """Crop the image (shape [C, D, H, W] or [C, H, W]) based on bounding box
    """
    def __init__(self, params):
        """
        start (None or tuple/list): The start index along each spatial axis.
            if None, calculate the start index automatically so that 
            the cropped region is centered at the non-zero region.
        output_size (None or tuple/list): Desired spatial output size.
            if None, set it as the size of bounding box of non-zero region 
        """
        self.start       = params['CropWithBoundingBox_start'.lower()]
        self.output_size = params['CropWithBoundingBox_output_size'.lower()]
        self.inverse     = params.get('CropWithBoundingBox_inverse'.lower(), True)
        self.task = params['task']
        
    def get_crop_param(self, sample):
        image = sample['image']
        input_shape = sample['image'].shape
        input_dim   = len(input_shape) - 1
        bb_min, bb_max = get_ND_bounding_box(image)
        bb_min, bb_max = bb_min[1:], bb_max[1:]
        if(self.start is None):
            if(self.output_size is None):
                crop_min, crop_max = bb_min, bb_max
            else:
                assert(len(self.output_size) == input_dim)
                crop_min = [int((bb_min[i] + bb_max[i] + 1)/2) - int(self.output_size[i]/2) \
                    for i in range(input_dim)]
                crop_min = [max(0, crop_min[i]) for i in range(input_dim)]
                crop_max = [crop_min[i] + self.output_size[i] for i in range(input_dim)]
        else:
            assert(len(self.start) == input_dim)
            crop_min = self.start
            if(self.output_size is None):
                assert(len(self.output_size) == input_dim)
                crop_max = [crop_min[i] + bb_max[i] - bb_min[i] \
                    for i in range(input_dim)]
            else:
                crop_max =  [crop_min[i] + self.output_size[i] for i in range(input_dim)]
        crop_min = [0] + crop_min
        crop_max = list(input_shape[0:1]) + crop_max
        sample['CropWithBoundingBox_Param'] = json.dumps((input_shape, crop_min, crop_max))   
        print("for crop", crop_min, crop_max)
        return sample, crop_min, crop_max

    def get_param_for_inverse_transform(self, sample):
        if(isinstance(sample['CropWithBoundingBox_Param'], list) or \
            isinstance(sample['CropWithBoundingBox_Param'], tuple)):
            params = json.loads(sample['CropWithBoundingBox_Param'][0]) 
        else:
            params = json.loads(sample['CropWithBoundingBox_Param']) 
        return params
        

class RandomCrop(CenterCrop):
    """Randomly crop the input image (shape [C, D, H, W] or [C, H, W]) 
    """
    def __init__(self, params):
        """
        output_size (tuple or list): Desired output size [D, H, W] or [H, W].
            the output channel is the same as the input channel.
            If D is None for 3D images, the z-axis is not cropped
        foreground_focus (bool): If true, allow crop around the foreground.
        foreground_ratio (float): Specifying the probability of foreground 
            focus cropping when foreground_focus is true.
        mask_label (None, or tuple / list): Specifying the foreground labels for foreground 
            focus cropping
        """
        # super(RandomCrop, self).__init__(params)
        self.output_size = params['RandomCrop_output_size'.lower()]
        self.fg_focus    = params.get('RandomCrop_foreground_focus'.lower(), False)
        self.fg_ratio    = params.get('RandomCrop_foreground_ratio'.lower(), 0.5)
        self.mask_label  = params.get('RandomCrop_mask_label'.lower(), [1])
        self.inverse     = params.get('RandomCrop_inverse'.lower(), True)
        self.task        = params['Task'.lower()]
        assert isinstance(self.output_size, (list, tuple))
        if(self.mask_label is not None):
            assert isinstance(self.mask_label, (list, tuple))

    def get_crop_param(self, sample):
        image       = sample['image']
        input_shape = image.shape
        input_dim   = len(input_shape) - 1
        assert(input_dim == len(self.output_size))
        temp_output_size = self.output_size
        if(input_dim == 3 and self.output_size[0] is None):
            # note that output size is [D, H, W] and input is [C, D, H, W]
            temp_output_size = [input_shape[1]] + self.output_size[1:]

        crop_margin = [input_shape[i + 1] - temp_output_size[i]\
            for i in range(input_dim)]
        crop_min = [0 if item == 0 else random.randint(0, item) for item in crop_margin]
        if(self.fg_focus and random.random() < self.fg_ratio):
            label = sample['label']
            mask  = np.zeros_like(label)
            for temp_lab in self.mask_label:
                mask = np.maximum(mask, label == temp_lab)
            if(mask.sum() == 0):
                bb_min = [0] * (input_dim + 1)
                bb_max = mask.shape
            else:
                bb_min, bb_max = get_ND_bounding_box(mask)
            bb_min, bb_max = bb_min[1:], bb_max[1:]
            crop_min = [random.randint(bb_min[i], bb_max[i]) - int(temp_output_size[i]/2) \
                for i in range(input_dim)]
            crop_min = [max(0, item) for item in crop_min]
            crop_min = [min(crop_min[i], input_shape[i+1] - temp_output_size[i]) \
                for i in range(input_dim)]

        crop_max = [crop_min[i] + temp_output_size[i] \
            for i in range(input_dim)]
        crop_min = [0] + crop_min
        crop_max = list(input_shape[0:1]) + crop_max
        sample['RandomCrop_Param'] = json.dumps((input_shape, crop_min, crop_max))
        return sample, crop_min, crop_max

    def get_param_for_inverse_transform(self, sample):
        if(isinstance(sample['RandomCrop_Param'], list) or \
            isinstance(sample['RandomCrop_Param'], tuple)):
            params = json.loads(sample['RandomCrop_Param'][0]) 
        else:
            params = json.loads(sample['RandomCrop_Param']) 
        return params

class RandomResizedCrop(CenterCrop):
    """Randomly crop the input image (shape [C, H, W])
       Only 2D images are supported 
    """
    def __init__(self, params):
        """
        output_size (tuple or list): Desired output size [H, W].
            the output channel is the same as the input channel.
        scale (tuple or list): range of scale, e.g. (0.08, 1.0)
        ratio (tuple or list): range of aspect ratio, e.g. (0.75, 1.33)
        """
        self.output_size = params['RandomResizedCrop_output_size'.lower()]
        self.scale       = params['RandomResizedCrop_scale'.lower()]
        self.ratio       = params['RandomResizedCrop_ratio'.lower()]
        self.inverse     = params.get('RandomResizedCrop_inverse'.lower(), True)
        self.task        = params['Task'.lower()]
        assert isinstance(self.output_size, (list, tuple))
        assert isinstance(self.scale, (list, tuple))
        assert isinstance(self.ratio, (list, tuple))
        
    def get_crop_param(self, sample):
        image = sample['image']
        input_shape = image.shape
        input_dim   = len(input_shape) - 1
        assert(input_dim == 2)
        assert(input_dim == len(self.output_size))
        
        scale = self.scale[0] + random.random()*(self.scale[1] - self.scale[0])
        ratio = self.ratio[0] + random.random()*(self.ratio[1] - self.ratio[0])
        crop_w = input_shape[-1] * scale 
        crop_h = crop_w * ratio
        crop_h = min(crop_h, input_shape[-2])
        output_shape = [int(crop_h), int(crop_w)]

        crop_margin = [input_shape[i + 1] - output_shape[i]\
            for i in range(input_dim)]
        crop_min = [random.randint(0, item) for item in crop_margin]
        crop_max = [crop_min[i] + output_shape[i] \
            for i in range(input_dim)]
        crop_min = [0] + crop_min
        crop_max = list(input_shape[0:1]) + crop_max
        sample['RandomResizedCrop_Param'] = json.dumps((input_shape, crop_min, crop_max))
        return sample, crop_min, crop_max

    def __call__(self, sample):
        image = sample['image']
        input_shape = image.shape
        input_dim   = len(input_shape) - 1
        sample, crop_min, crop_max = self.get_crop_param(sample)

        image_t = crop_ND_volume_with_bounding_box(image, crop_min, crop_max)
        crp_shape = image_t.shape
        scale = [(self.output_size[i] + 0.0)/crp_shape[1:][i] for i in range(input_dim)]
        scale = [1.0] + scale
        image_t = ndimage.interpolation.zoom(image_t, scale, order = 1)
        sample['image'] = image_t
        
        if('label' in sample and self.task == 'segmentation'):
            label = sample['label']
            crop_max[0] = label.shape[0]
            label = crop_ND_volume_with_bounding_box(label, crop_min, crop_max)
            label = ndimage.interpolation.zoom(label, scale, order = 0)
            sample['label'] = label
        if('pixel_weight' in sample and self.task == 'segmentation'):
            weight = sample['pixel_weight']
            crop_max[0] = weight.shape[0]
            weight = crop_ND_volume_with_bounding_box(weight, crop_min, crop_max)
            weight = ndimage.interpolation.zoom(weight, scale, order = 1)
            sample['pixel_weight'] = weight
        return sample

    def inverse_transform_for_prediction(self, sample):
        """
        not implemented
        """
        raise(ValueError("not implemented"))


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn
from pymic.loss.seg.util import reshape_tensor_to_2D, get_classwise_dice

class DiceLoss(nn.Module):
    def __init__(self, params = None):
        super(DiceLoss, self).__init__()
        if(params is None):
            self.softmax = True
        else:
            self.softmax = params.get('loss_softmax', True)

    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        soft_y  = loss_input_dict['ground_truth']
        
        if(isinstance(predict, (list, tuple))):
            predict = predict[0]
        if(self.softmax):
            predict = nn.Softmax(dim = 1)(predict)
        predict = reshape_tensor_to_2D(predict)
        soft_y  = reshape_tensor_to_2D(soft_y) 
        dice_score = get_classwise_dice(predict, soft_y)
        dice_loss  = 1.0 - dice_score.mean()
        return dice_loss

class FocalDiceLoss(nn.Module):
    """
    focal Dice according to the following paper:
    Pei Wang and Albert C. S. Chung, Focal Dice Loss and Image Dilation for 
    Brain Tumor Segmentation, 2018
    """
    def __init__(self, params = None):
        super(FocalDiceLoss, self).__init__()
        self.softmax = params.get('loss_softmax', True)
        self.beta = params['FocalDiceLoss_beta'.lower()] #beta should be >=1.0

    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        soft_y  = loss_input_dict['ground_truth']

        if(isinstance(predict, (list, tuple))):
            predict = predict[0]
        if(self.softmax):
            predict = nn.Softmax(dim = 1)(predict)
        predict = reshape_tensor_to_2D(predict)
        soft_y  = reshape_tensor_to_2D(soft_y) 

        dice_score = get_classwise_dice(predict, soft_y)
        dice_score = torch.pow(dice_score, 1.0 / self.beta)
        dice_loss  = 1.0 - dice_score.mean()
        return dice_loss

class NoiseRobustDiceLoss(nn.Module):
    """
    Noise-robust Dice loss according to the following paper. 
        G. Wang et al. A Noise-Robust Framework for Automatic Segmentation of COVID-19 
        Pneumonia Lesions From CT Images, IEEE TMI, 2020. 
        https://doi.org/10.1109/TMI.2020.3000314
    """
    def __init__(self, params):
        super(NoiseRobustDiceLoss, self).__init__()
        self.softmax = params.get('loss_softmax', True)
        self.gamma = params['NoiseRobustDiceLoss_gamma'.lower()]

    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        soft_y  = loss_input_dict['ground_truth']

        if(isinstance(predict, (list, tuple))):
            predict = predict[0]
        if(self.softmax):
            predict = nn.Softmax(dim = 1)(predict)
        predict = reshape_tensor_to_2D(predict)
        soft_y  = reshape_tensor_to_2D(soft_y) 

        numerator = torch.abs(predict - soft_y)
        numerator = torch.pow(numerator, self.gamma)
        denominator = predict + soft_y 
        numer_sum = torch.sum(numerator,  dim = 0)
        denom_sum = torch.sum(denominator,  dim = 0)
        loss_vector = numer_sum / (denom_sum + 1e-5)
        loss = torch.mean(loss_vector)   
        return loss



# -*- coding: utf-8 -*-
"""
Author: Guotai Wang
Date:   12 June, 2020
Implementation of of COPLENet for COVID-19 pneumonia lesion segmentation from CT images.
Reference: 
    G. Wang et al. A Noise-robust Framework for Automatic Segmentation of COVID-19 Pneumonia Lesions 
    from CT Images. IEEE Transactions on Medical Imaging, 39(8),2020:2653-2663. DOI:10.1109/TMI.2020.3000314.
"""

from __future__ import print_function, division
import torch
import torch.nn as nn

class ConvLayer(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size = 1):
        super(ConvLayer, self).__init__()
        padding = int((kernel_size - 1) / 2)
        self.conv = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, padding=padding),
            nn.BatchNorm2d(out_channels),
            nn.LeakyReLU()
        )
       
    def forward(self, x):
        return self.conv(x)

class SEBlock(nn.Module):
    def __init__(self, in_channels, r):
        super(SEBlock, self).__init__()

        redu_chns = int(in_channels / r)
        self.se_layers = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_channels, redu_chns, kernel_size=1, padding=0),
            nn.LeakyReLU(),
            nn.Conv2d(redu_chns, in_channels, kernel_size=1, padding=0),
            nn.ReLU())
        
    def forward(self, x):
        f = self.se_layers(x)
        return f*x + x

class ASPPBlock(nn.Module):
    def __init__(self,in_channels, out_channels_list, kernel_size_list, dilation_list):
        super(ASPPBlock, self).__init__()
        self.conv_num = len(out_channels_list)
        assert(self.conv_num == 4)
        assert(self.conv_num == len(kernel_size_list) and self.conv_num == len(dilation_list))
        pad0 = int((kernel_size_list[0] - 1) / 2 * dilation_list[0])
        pad1 = int((kernel_size_list[1] - 1) / 2 * dilation_list[1])
        pad2 = int((kernel_size_list[2] - 1) / 2 * dilation_list[2])
        pad3 = int((kernel_size_list[3] - 1) / 2 * dilation_list[3])
        self.conv_1 = nn.Conv2d(in_channels, out_channels_list[0], kernel_size = kernel_size_list[0], 
                    dilation = dilation_list[0], padding = pad0 )
        self.conv_2 = nn.Conv2d(in_channels, out_channels_list[1], kernel_size = kernel_size_list[1], 
                    dilation = dilation_list[1], padding = pad1 )
        self.conv_3 = nn.Conv2d(in_channels, out_channels_list[2], kernel_size = kernel_size_list[2], 
                    dilation = dilation_list[2], padding = pad2 )
        self.conv_4 = nn.Conv2d(in_channels, out_channels_list[3], kernel_size = kernel_size_list[3], 
                    dilation = dilation_list[3], padding = pad3 )

        out_channels = out_channels_list[0] + out_channels_list[1] + out_channels_list[2] + out_channels_list[3] 
        self.conv_1x1 = nn.Sequential(
            nn.Conv2d(out_channels, out_channels, kernel_size=1, padding=0),
            nn.BatchNorm2d(out_channels),
            nn.LeakyReLU())
       
    def forward(self, x):
        x1 = self.conv_1(x)
        x2 = self.conv_2(x)
        x3 = self.conv_3(x)
        x4 = self.conv_4(x)

        y = torch.cat([x1, x2, x3, x4], dim=1)
        y = self.conv_1x1(y)
        return y

class ConvBNActBlock(nn.Module):
    """Two convolution layers with batch norm, leaky relu, dropout and SE block"""
    def __init__(self,in_channels, out_channels, dropout_p):
        super(ConvBNActBlock, self).__init__()
        self.conv_conv = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.LeakyReLU(),
            nn.Dropout(dropout_p),
            nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.LeakyReLU(),
            SEBlock(out_channels, 2)
        )
       
    def forward(self, x):
        return self.conv_conv(x)

class DownBlock(nn.Module):
    """Downsampling by a concantenation of max-pool and avg-pool, followed by ConvBNActBlock
    """
    def __init__(self, in_channels, out_channels, dropout_p):
        super(DownBlock, self).__init__()
        self.maxpool = nn.MaxPool2d(2)
        self.avgpool = nn.AvgPool2d(2)
        self.conv = ConvBNActBlock(2 * in_channels, out_channels, dropout_p)
        
    def forward(self, x):
        x_max = self.maxpool(x)
        x_avg = self.avgpool(x)
        x_cat = torch.cat([x_max, x_avg], dim=1)
        y = self.conv(x_cat)
        return y + x_cat

class UpBlock(nn.Module):
    """Upssampling followed by ConvBNActBlock"""
    def __init__(self, in_channels1, in_channels2, out_channels, 
                 bilinear=True, dropout_p = 0.5):
        super(UpBlock, self).__init__()
        self.bilinear = bilinear
        if bilinear:
            self.conv1x1 = nn.Conv2d(in_channels1, in_channels2, kernel_size = 1)
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
        else:
            self.up = nn.ConvTranspose2d(in_channels1, in_channels2, kernel_size=2, stride=2)
        self.conv = ConvBNActBlock(in_channels2 * 2, out_channels, dropout_p)

    def forward(self, x1, x2):
        if self.bilinear:
            x1 = self.conv1x1(x1)
        x1    = self.up(x1)
        x_cat = torch.cat([x2, x1], dim=1)
        y     = self.conv(x_cat)
        return y + x_cat

class COPLENet(nn.Module):
    def __init__(self, params):
        super(COPLENet, self).__init__()
        self.params    = params
        self.in_chns   = self.params['in_chns']
        self.ft_chns   = self.params['feature_chns']
        self.n_class   = self.params['class_num']
        self.bilinear  = self.params['bilinear']
        self.dropout   = self.params['dropout']
        assert(len(self.ft_chns) == 5)

        f0_half = int(self.ft_chns[0] / 2)
        f1_half = int(self.ft_chns[1] / 2)
        f2_half = int(self.ft_chns[2] / 2)
        f3_half = int(self.ft_chns[3] / 2)
        self.in_conv= ConvBNActBlock(self.in_chns, self.ft_chns[0], self.dropout[0])
        self.down1  = DownBlock(self.ft_chns[0], self.ft_chns[1], self.dropout[1])
        self.down2  = DownBlock(self.ft_chns[1], self.ft_chns[2], self.dropout[2])
        self.down3  = DownBlock(self.ft_chns[2], self.ft_chns[3], self.dropout[3])
        self.down4  = DownBlock(self.ft_chns[3], self.ft_chns[4], self.dropout[4])
        
        self.bridge0= ConvLayer(self.ft_chns[0], f0_half)
        self.bridge1= ConvLayer(self.ft_chns[1], f1_half)
        self.bridge2= ConvLayer(self.ft_chns[2], f2_half)
        self.bridge3= ConvLayer(self.ft_chns[3], f3_half)

        self.up1    = UpBlock(self.ft_chns[4], f3_half, self.ft_chns[3], dropout_p = self.dropout[3])
        self.up2    = UpBlock(self.ft_chns[3], f2_half, self.ft_chns[2], dropout_p = self.dropout[2])
        self.up3    = UpBlock(self.ft_chns[2], f1_half, self.ft_chns[1], dropout_p = self.dropout[1])
        self.up4    = UpBlock(self.ft_chns[1], f0_half, self.ft_chns[0], dropout_p = self.dropout[0])

        f4 = self.ft_chns[4]
        aspp_chns = [int(f4 / 4), int(f4 / 4), int(f4 / 4), int(f4 / 4)]
        aspp_knls = [1, 3, 3, 3]
        aspp_dila = [1, 2, 4, 6]
        self.aspp = ASPPBlock(f4, aspp_chns, aspp_knls, aspp_dila)
        
            
        self.out_conv = nn.Conv2d(self.ft_chns[0], self.n_class,  
            kernel_size = 3, padding = 1)

    def forward(self, x):
        x_shape = list(x.shape)
        if(len(x_shape) == 5):
          [N, C, D, H, W] = x_shape
          new_shape = [N*D, C, H, W]
          x = torch.transpose(x, 1, 2)
          x = torch.reshape(x, new_shape)
        x0  = self.in_conv(x)
        x0b = self.bridge0(x0)
        x1  = self.down1(x0)
        x1b = self.bridge1(x1)
        x2  = self.down2(x1)
        x2b = self.bridge2(x2)
        x3  = self.down3(x2)
        x3b = self.bridge3(x3)
        x4  = self.down4(x3)
        x4  = self.aspp(x4) 

        x = self.up1(x4, x3b)
        x = self.up2(x, x2b)
        x = self.up3(x, x1b)
        x = self.up4(x, x0b)
        output = self.out_conv(x)

        if(len(x_shape) == 5):
            new_shape = [N, D] + list(output.shape)[1:]
            output = torch.reshape(output, new_shape)
            output = torch.transpose(output, 1, 2)
        return output



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn
import numpy as np 
from pymic.loss.seg.util import reshape_tensor_to_2D

class EntropyLoss(nn.Module):
    """
    Minimize the entropy for each pixel
    """
    def __init__(self, params = None):
        super(EntropyLoss, self).__init__()
        if(params is None):
            self.softmax = True
        else:
            self.softmax = params.get('loss_softmax', True)
        
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']

        if(isinstance(predict, (list, tuple))):
            predict = predict[0]
        if(self.softmax):
            predict = nn.Softmax(dim = 1)(predict)

        # for numeric stability
        predict = predict * 0.999 + 5e-4
        C = list(predict.shape)[1]
        entropy = torch.sum(-predict*torch.log(predict), dim=1) / np.log(C)
        avg_ent = torch.mean(entropy)
        return avg_ent
    
class TotalVariationLoss(nn.Module):
    """
    Minimize the total variation of a segmentation
    """
    def __init__(self, params = None):
        super(TotalVariationLoss, self).__init__()
        if(params is None):
            self.softmax = True
        else:
            self.softmax = params.get('loss_softmax', True)
        
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']

        if(isinstance(predict, (list, tuple))):
            predict = predict[0]
        if(self.softmax):
            predict = nn.Softmax(dim = 1)(predict)

        # for numeric stability
        predict = predict * 0.999 + 5e-4
        dim = list(predict.shape)[2:]
        if(dim == 2):
            pred_min = -1 * nn.functional.max_pool2d(-1*predict, (3, 3), 1, 1)
            pred_max = nn.functional.max_pool2d(pred_min, (3, 3), 1, 1)
        else:
            pred_min = -1 * nn.functional.max_pool3d(-1*predict, (3, 3, 3), 1, 1)
            pred_max = nn.functional.max_pool3d(pred_min, (3, 3, 3), 1, 1)            
        contour = torch.relu(pred_max - pred_min)
        length  = torch.mean(contour)
        return length



# -*- coding: utf-8 -*-
from __future__ import print_function, division
import numpy as np

"""Functions for ramping hyperparameters up or down

Each function takes the current training step or epoch, and the
ramp length (maximal step or epoch), and returns a multiplier between
0 and 1.
"""


def get_rampup_ratio(i, start, end, mode = "linear"):
    if( i < start):
        rampup = 0.0
    elif(i > end):
        rampup = 1.0
    elif(mode == "linear"):
        rampup = (i - start) / (end - start)
    elif(mode == "sigmoid"):
        phase  = 1.0 - (i - start) / (end - start)
        rampup = float(np.exp(-5.0 * phase * phase))
    return rampup


def cosine_rampdown(i, start, end):
    """Cosine rampdown from https://arxiv.org/abs/1608.03983"""
    i = np.clip(i, 0.0, length)
    return float(.5 * (np.cos(np.pi * i / length) + 1))


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn

class NLLLoss(nn.Module):
    def __init__(self, params):
        super(NLLLoss, self).__init__()
        self.nll_loss = nn.NLLLoss()
    
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        labels  = loss_input_dict['ground_truth']
        logsoft = nn.LogSoftmax(dim = 1)
        predict = logsoft(predict)
        loss = self.nll_loss(predict, labels)
        return loss



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn

class CrossEntropyLoss(nn.Module):
    """
    Standard Softmax-based  CE loss
    Args:
    predict has a shape of [N, C] where C is the class number
    labels  has a shape of [N]

    note that predict is the digit output of a network, before using softmax
    """
    def __init__(self, params):
        super(CrossEntropyLoss, self).__init__()
        self.ce_loss = nn.CrossEntropyLoss()
    
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        labels  = loss_input_dict['ground_truth']
        loss = self.ce_loss(predict, labels)
        return loss

class SigmoidCELoss(nn.Module):
    """
    Sigmoid-based CE loss
    Args:
    predict has a shape of [N, C] where C is the class number
    labels  has a shape of [N, C] with binary values
    note that predict is the digit output of a network, before using sigmoid."""
    def __init__(self, params):
        super(SigmoidCELoss, self).__init__()
    
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        labels  = loss_input_dict['ground_truth']
        predict = nn.Sigmoid()(predict) * 0.999 + 5e-4
        loss = - labels * torch.log(predict) - (1 - labels) * torch.log( 1 - predict)
        loss = loss.mean()
        return loss


# -*- coding: utf-8 -*-
from __future__ import print_function, division
import logging
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from torch.optim import lr_scheduler
from pymic.loss.seg.util import get_soft_label
from pymic.loss.seg.util import reshape_prediction_and_ground_truth
from pymic.loss.seg.util import get_classwise_dice
from pymic.net_run_ssl.ssl_abstract import SSLSegAgent
from pymic.util.ramps import get_rampup_ratio

def softmax_mse_loss(inputs, targets, conf_mask=False, threshold=None, use_softmax=False):
    assert inputs.requires_grad == True and targets.requires_grad == False
    assert inputs.size() == targets.size() # (batch_size * num_classes * H * W)
    inputs = F.softmax(inputs, dim=1)
    if use_softmax:
        targets = F.softmax(targets, dim=1)

    if conf_mask:
        loss_mat = F.mse_loss(inputs, targets, reduction='none')
        mask = (targets.max(1)[0] > threshold)
        loss_mat = loss_mat[mask.unsqueeze(1).expand_as(loss_mat)]
        if loss_mat.shape.numel() == 0: loss_mat = torch.tensor([0.]).to(inputs.device)
        return loss_mat.mean()
    else:
        return F.mse_loss(inputs, targets, reduction='mean') # take the mean over the batch_size


def softmax_kl_loss(inputs, targets, conf_mask=False, threshold=None, use_softmax=False):
    assert inputs.requires_grad == True and targets.requires_grad == False
    assert inputs.size() == targets.size()
    input_log_softmax = F.log_softmax(inputs, dim=1)
    if use_softmax:
        targets = F.softmax(targets, dim=1)
    
    if conf_mask:
        loss_mat = F.kl_div(input_log_softmax, targets, reduction='none')
        mask = (targets.max(1)[0] > threshold)
        loss_mat = loss_mat[mask.unsqueeze(1).expand_as(loss_mat)]
        if loss_mat.shape.numel() == 0: loss_mat = torch.tensor([0.]).to(inputs.device)
        return loss_mat.sum() / mask.shape.numel()
    else:
        return F.kl_div(input_log_softmax, targets, reduction='mean')


def softmax_js_loss(inputs, targets, **_):
    assert inputs.requires_grad == True and targets.requires_grad == False
    assert inputs.size() == targets.size()
    epsilon = 1e-5

    M = (F.softmax(inputs, dim=1) + targets) * 0.5
    kl1 = F.kl_div(F.log_softmax(inputs, dim=1), M, reduction='mean')
    kl2 = F.kl_div(torch.log(targets+epsilon), M, reduction='mean')
    return (kl1 + kl2) * 0.5

unsup_loss_dict = {"MSE": softmax_mse_loss,
   "KL":softmax_kl_loss,
   "JS":softmax_js_loss}

class SSLCCT(SSLSegAgent):
    """
    Cross-Consistency Training according to the following paper:
        Yassine Ouali, Celine Hudelot and Myriam Tami:
        Semi-Supervised Semantic Segmentation With Cross-Consistency Training. 
        CVPR 2020.
        https://arxiv.org/abs/2003.09005          
    Code adapted from: https://github.com/yassouali/CCT
    """
    def training(self):
        class_num   = self.config['network']['class_num']
        iter_valid  = self.config['training']['iter_valid']
        ssl_cfg     = self.config['semi_supervised_learning']
        iter_max     = self.config['training']['iter_max']
        rampup_start = ssl_cfg.get('rampup_start', 0)
        rampup_end   = ssl_cfg.get('rampup_end', iter_max)
        unsup_loss_name = ssl_cfg.get('unsupervised_loss', "MSE")
        self.unsup_loss_f = unsup_loss_dict[unsup_loss_name]
        train_loss  = 0
        train_loss_sup = 0
        train_loss_reg = 0
        train_dice_list = []
        self.net.train()
        
        for it in range(iter_valid):
            try:
                data_lab = next(self.trainIter)
            except StopIteration:
                self.trainIter = iter(self.train_loader)
                data_lab = next(self.trainIter)
            try:
                data_unlab = next(self.trainIter_unlab)
            except StopIteration:
                self.trainIter_unlab = iter(self.train_loader_unlab)
                data_unlab = next(self.trainIter_unlab)

            # get the inputs
            x0   = self.convert_tensor_type(data_lab['image'])
            y0   = self.convert_tensor_type(data_lab['label_prob'])  
            x1   = self.convert_tensor_type(data_unlab['image'])
            inputs = torch.cat([x0, x1], dim = 0)               
            inputs, y0 = inputs.to(self.device), y0.to(self.device)
            
            # zero the parameter gradients
            self.optimizer.zero_grad()

            # forward pass
            output, aux_outputs = self.net(inputs)
            n0 = list(x0.shape)[0] 

            # get supervised loss
            p0 = output[:n0]
            loss_sup = self.get_loss_value(data_lab, p0, y0)

            # get regularization loss
            p1 = F.softmax(output[n0:].detach(), dim=1)
            p1_aux = [aux_out[n0:] for aux_out in aux_outputs]
            loss_reg = 0.0
            for p1_auxi in p1_aux:
                loss_reg += self.unsup_loss_f( p1_auxi, p1, use_softmax = True)
            loss_reg = loss_reg / len(p1_aux)
            
            rampup_ratio = get_rampup_ratio(self.glob_it, rampup_start, rampup_end, "sigmoid")
            regular_w = ssl_cfg.get('regularize_w', 0.1) * rampup_ratio
            loss = loss_sup + regular_w*loss_reg

            loss.backward()
            self.optimizer.step()
            if(self.scheduler is not None and \
                not isinstance(self.scheduler, lr_scheduler.ReduceLROnPlateau)):
                self.scheduler.step()
            train_loss = train_loss + loss.item()
            train_loss_sup = train_loss_sup + loss_sup.item()
            train_loss_reg = train_loss_reg + loss_reg.item() 
            # get dice evaluation for each class in annotated images
            if(isinstance(p0, tuple) or isinstance(p0, list)):
                p0 = p0[0] 
            p0_argmax = torch.argmax(p0, dim = 1, keepdim = True)
            p0_soft   = get_soft_label(p0_argmax, class_num, self.tensor_type)
            p0_soft, y0 = reshape_prediction_and_ground_truth(p0_soft, y0) 
            dice_list   = get_classwise_dice(p0_soft, y0)
            train_dice_list.append(dice_list.cpu().numpy())
        train_avg_loss = train_loss / iter_valid
        train_avg_loss_sup = train_loss_sup / iter_valid
        train_avg_loss_reg = train_loss_reg / iter_valid
        train_cls_dice = np.asarray(train_dice_list).mean(axis = 0)
        train_avg_dice = train_cls_dice.mean()

        train_scalers = {'loss': train_avg_loss, 'loss_sup':train_avg_loss_sup,
            'loss_reg':train_avg_loss_reg, 'regular_w':regular_w,
            'avg_dice':train_avg_dice,     'class_dice': train_cls_dice}
        return train_scalers



import json
import math
import random
import numpy as np
from scipy import ndimage
from pymic.transform.abstract_transform import AbstractTransform
from pymic.util.image_process import *


class NormalizeWithMeanStd(AbstractTransform):
    """Nomralize the image (shape [C, D, H, W] or [C, H, W]) with mean and std for given channels
    """
    def __init__(self, params):
        """
        :param chanels: (None or tuple/list) the indices of channels to be noramlized.
        :param mean: (None or tuple/list): The mean values along each channel.
        :param  std : (None or tuple/list): The std values along each channel.
            When mean and std are not provided, calculate them from the entire image
            region or the non-positive region.
        :param ignore_non_positive: (bool) Only used when mean and std are not given. 
            Use positive region to calculate mean and std, and set non-positive region to random.  
        :param inverse: (bool) Whether inverse transform is needed or not.
        """
        super(NormalizeWithMeanStd, self).__init__(params)
        self.chns = params['NormalizeWithMeanStd_channels'.lower()]
        self.mean = params.get('NormalizeWithMeanStd_mean'.lower(), None)
        self.std  = params.get('NormalizeWithMeanStd_std'.lower(), None)
        self.ingore_np = params.get('NormalizeWithMeanStd_ignore_non_positive'.lower(), False)
        self.inverse = params.get('NormalizeWithMeanStd_inverse'.lower(), False)

    def __call__(self, sample):
        image= sample['image']
        chns = self.chns if self.chns is not None else range(image.shape[0])
        if(self.mean is None):
            self.mean = [None] * len(chns)
            self.std  = [None] * len(chns)

        for i in range(len(chns)):
            chn = chns[i]
            chn_mean, chn_std = self.mean[i], self.std[i]
            if(chn_mean is None):
                if(self.ingore_np):
                    pixels = image[chn][image[chn] > 0]
                    chn_mean, chn_std = pixels.mean(), pixels.std() 
                else:
                    chn_mean, chn_std = image[chn].mean(), image[chn].std()
    
            chn_norm = (image[chn] - chn_mean)/chn_std

            if(self.ingore_np):
                chn_random = np.random.normal(0, 1, size = chn_norm.shape)
                chn_norm[image[chn] <= 0] = chn_random[image[chn] <= 0]
            image[chn] = chn_norm
        sample['image'] = image
        return sample


class NormalizeWithMinMax(AbstractTransform):
    """Nomralize the image (shape [C, D, H, W] or [C, H, W]) with min and max for given channels
    """
    def __init__(self, params):
        """
        :param chanels: (None or tuple/list) the indices of channels to be noramlized.
        :param threshold_lower: (tuple/list/None) The lower threshold value along each channel.
        :param threshold_upper: (typle/list/None) The upper threshold value along each channel.
        :param inverse: (bool) Whether inverse transform is needed or not.
        """
        super(NormalizeWithMinMax, self).__init__(params)
        self.chns = params['NormalizeWithMinMax_channels'.lower()]
        self.thred_lower = params['NormalizeWithMinMax_threshold_lower'.lower()]
        self.thred_upper = params['NormalizeWithMinMax_threshold_upper'.lower()]
        self.inverse = params.get('NormalizeWithMinMax_inverse'.lower(), False)

    def __call__(self, sample):
        image= sample['image']
        chns = self.chns if self.chns is not None else range(image.shape[0])
        for i in range(len(chns)):
            chn = chns[i]
            img_chn = image[chn]
            v0, v1 = img_chn.min(), img_chn.max()
            if(self.thred_lower is not None) and  (self.thred_lower[i] is not None):
                v0 = self.thred_lower[i]
            if(self.thred_upper is not None) and  (self.thred_upper[i] is not None):
                v1 = self.thred_upper[i]

            img_chn[img_chn < v0] = v0
            img_chn[img_chn > v1] = v1
            img_chn = (img_chn - v0) / (v1 - v0)
            image[chn] = img_chn
        sample['image'] = image
        return sample

class NormalizeWithPercentiles(AbstractTransform):
    """Nomralize the image (shape [C, D, H, W] or [C, H, W]) with percentiles for given channels
    """
    def __init__(self, params):
        """
        :param chanels: (None or tuple/list) the indices of channels to be noramlized.
        :param percentile_lower: (tuple/list/None) The lower percentile along each channel.
        :param percentile_upper: (typle/list/None) The upper percentile along each channel.
        :param inverse: (bool) Whether inverse transform is needed or not.
        """
        super(NormalizeWithPercentiles, self).__init__(params)
        self.chns = params['NormalizeWithPercentiles_channels'.lower()]
        self.percent_lower = params['NormalizeWithPercentiles_percentile_lower'.lower()]
        self.percent_upper = params['NormalizeWithPercentiles_percentile_upper'.lower()]
        self.inverse = params.get('NormalizeWithPercentiles_inverse'.lower(), False)

    def __call__(self, sample):
        image= sample['image']
        chns = self.chns if self.chns is not None else range(image.shape[0])
        for i in range(len(chns)):
            chn = chns[i]
            img_chn = image[chn]
            v0 = np.percentile(img_chn, self.percent_lower)
            v1 = np.percentile(img_chn, self.percent_upper)

            img_chn[img_chn < v0] = v0
            img_chn[img_chn > v1] = v1
            img_chn = (img_chn - v0) / (v1 - v0)
            image[chn] = img_chn
        sample['image'] = image
        return sample


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import copy
import csv
import logging
import time
import torch
from torchvision import transforms
import numpy as np
import torch.nn as nn
from datetime import datetime
from tensorboardX import SummaryWriter
from pymic.io.nifty_dataset import ClassificationDataset
from pymic.loss.loss_dict_cls import PyMICClsLossDict
from pymic.net.net_dict_cls import TorchClsNetDict
from pymic.transform.trans_dict import TransformDict
from pymic.net_run.agent_abstract import NetRunAgent
import warnings
warnings.filterwarnings('ignore', '.*output shape of zoom.*')

class ClassificationAgent(NetRunAgent):
    def __init__(self, config, stage = 'train'):
        super(ClassificationAgent, self).__init__(config, stage)
        self.transform_dict  = TransformDict
        assert(self.task_type in  ["cls", "cls_nexcl"])

    def get_stage_dataset_from_config(self, stage):
        assert(stage in ['train', 'valid', 'test'])
        root_dir  = self.config['dataset']['root_dir']
        modal_num = self.config['dataset']['modal_num']

        if(stage == "train"):
            transform_names = self.config['dataset']['train_transform']
        elif(stage == "valid"):
            transform_names = self.config['dataset']['valid_transform']
        elif(stage == "test"):
            transform_names = self.config['dataset']['test_transform']
        else:
            raise ValueError("Incorrect value for stage: {0:}".format(stage))
        self.transform_list  = []
        if(transform_names is None or len(transform_names) == 0):
            data_transform = None 
        else:
            transform_param = self.config['dataset']
            transform_param['task'] = 'classification' 
            for name in transform_names:
                if(name not in self.transform_dict):
                    raise(ValueError("Undefined transform {0:}".format(name))) 
                one_transform = self.transform_dict[name](transform_param)
                self.transform_list.append(one_transform)
            data_transform = transforms.Compose(self.transform_list)

        csv_file = self.config['dataset'].get(stage + '_csv', None)
        class_num = self.config['network']['class_num']
        dataset  = ClassificationDataset(root_dir=root_dir,
                                csv_file  = csv_file,
                                modal_num = modal_num,
                                class_num = class_num,
                                with_label= not (stage == 'test'),
                                transform = data_transform )
        return dataset

    def create_network(self):
        if(self.net is None):
            net_name = self.config['network']['net_type']
            if(net_name not in TorchClsNetDict):
                raise ValueError("Undefined network {0:}".format(net_name))
            self.net = TorchClsNetDict[net_name](self.config['network'])
        if(self.tensor_type == 'float'):
            self.net.float()
        else:
            self.net.double()
        param_number = sum(p.numel() for p in self.net.parameters() if p.requires_grad)
        logging.info('parameter number {0:}'.format(param_number))

    def get_parameters_to_update(self):
        params = self.net.get_parameters_to_update()
        return params

    def create_loss_calculator(self):
        if(self.loss_dict is None):
            self.loss_dict = PyMICClsLossDict
        loss_name = self.config['training']['loss_type']
        if(loss_name in self.loss_dict):
            self.loss_calculater = self.loss_dict[loss_name](self.config['training'])
        else:
            raise ValueError("Undefined loss function {0:}".format(loss_name))

    def get_loss_value(self, data, inputs, outputs, labels):
        loss_input_dict = {}
        loss_input_dict['prediction'] = outputs
        loss_input_dict['ground_truth'] = labels
        loss_value = self.loss_calculater(loss_input_dict)
        return loss_value
        
    def get_evaluation_score(self, outputs, labels):
        metrics = self.config['training'].get("evaluation_metric", "accuracy")
        if(metrics != "accuracy"): # default classification accuracy
            raise ValueError("Not implemeted for metric {0:}".format(metrics))
        if(self.task_type == "cls"):
            _, preds = torch.max(outputs, 1)
            consis= self.convert_tensor_type(preds ==  labels.data)
            score = torch.mean(consis)
        elif(self.task_type == "cls_nexcl"): #nonexclusive classification
            preds = self.convert_tensor_type(outputs > 0.5)
            consis= self.convert_tensor_type(preds ==  labels.data)
            score = torch.mean(consis) 
        return score

    def training(self):
        iter_valid   = self.config['training']['iter_valid']
        sample_num   = 0
        running_loss = 0
        running_score= 0
        self.net.train()
        for it in range(iter_valid):
            try:
                data = next(self.trainIter)
            except StopIteration:
                self.trainIter = iter(self.train_loader)
                data = next(self.trainIter)
            inputs = self.convert_tensor_type(data['image'])
            labels = data['label'].long()         
            inputs, labels = inputs.to(self.device), labels.to(self.device)
            # zero the parameter gradients
            self.optimizer.zero_grad()
            # forward + backward + optimize
            outputs = self.net(inputs)
            loss = self.get_loss_value(data, inputs, outputs, labels)
            loss.backward()
            self.optimizer.step()
            self.scheduler.step()
            
            # statistics
            sample_num   += labels.size(0)
            running_loss += loss.item() * labels.size(0)
            running_score+= self.get_evaluation_score(outputs, labels) * labels.size(0)

        avg_loss = running_loss / sample_num
        avg_score= running_score.double() / sample_num
        metrics =self.config['training'].get("evaluation_metric", "accuracy")
        train_scalers = {'loss': avg_loss, metrics: avg_score}
        return train_scalers

    def validation(self):
        validIter  = iter(self.valid_loader)
        sample_num   = 0
        running_loss = 0
        running_score= 0
        with torch.no_grad():
            self.net.eval()
            for data in validIter:
                inputs = self.convert_tensor_type(data['image'])
                labels = data['label'].long()             
                inputs, labels = inputs.to(self.device), labels.to(self.device)
                self.optimizer.zero_grad()
                # forward + backward + optimize
                outputs = self.net(inputs)
                loss = self.get_loss_value(data, inputs, outputs, labels)
                                
                # statistics
                sample_num   += labels.size(0)
                running_loss += loss.item() * labels.size(0)
                running_score+= self.get_evaluation_score(outputs, labels) * labels.size(0)

        avg_loss = running_loss / sample_num
        avg_score= running_score.double() / sample_num
        metrics =self.config['training'].get("evaluation_metric", "accuracy")
        valid_scalers = {'loss': avg_loss, metrics: avg_score}
        return valid_scalers

    def write_scalars(self, train_scalars, valid_scalars, glob_it):
        metrics =self.config['training'].get("evaluation_metric", "accuracy")
        loss_scalar ={'train':train_scalars['loss'], 'valid':valid_scalars['loss']}
        acc_scalar  ={'train':train_scalars[metrics],'valid':valid_scalars[metrics]}
        self.summ_writer.add_scalars('loss', loss_scalar, glob_it)
        self.summ_writer.add_scalars(metrics, acc_scalar, glob_it)
        
        logging.info("{0:} it {1:}".format(str(datetime.now())[:-7], glob_it))
        logging.info('train loss {0:.4f}, avg {1:} {2:.4f}'.format(
            train_scalars['loss'], metrics, train_scalars[metrics]))
        logging.info('valid loss {0:.4f}, avg {1:} {2:.4f}'.format(
            valid_scalars['loss'], metrics, valid_scalars[metrics])) 

    def train_valid(self):
        device_ids = self.config['training']['gpus']
        if(len(device_ids) > 1):
            self.device = torch.device("cuda:0")
            self.net = nn.DataParallel(self.net, device_ids = device_ids)
        else:
            self.device = torch.device("cuda:{0:}".format(device_ids[0]))
        self.net.to(self.device)

        ckpt_dir    = self.config['training']['ckpt_save_dir']
        if(ckpt_dir[-1] == "/"):
            ckpt_dir = ckpt_dir[:-1]
        ckpt_prefix = self.config['training'].get('ckpt_prefix', None)
        if(ckpt_prefix is None):
            ckpt_prefix = ckpt_dir.split('/')[-1]
        iter_start  = self.config['training']['iter_start']
        iter_max    = self.config['training']['iter_max']
        iter_valid  = self.config['training']['iter_valid']
        iter_save   = self.config['training']['iter_save']
        metrics     = self.config['training'].get("evaluation_metric", "accuracy")
        self.max_val_score  = 0.0
        self.max_val_it     = 0
        self.best_model_wts = None 
        self.checkpoint = None
        if(iter_start > 0):
            checkpoint_file = "{0:}/{1:}_{2:}.pt".format(ckpt_dir, ckpt_prefix, iter_start)
            self.checkpoint = torch.load(checkpoint_file, map_location = self.device)
            assert(self.checkpoint['iteration'] == iter_start)
            self.net.load_state_dict(self.checkpoint['model_state_dict'])
            self.max_val_score  = self.checkpoint.get('valid_pred', 0)
            self.max_val_it     = self.checkpoint['iteration']
            self.best_model_wts = self.checkpoint['model_state_dict']
        
        self.create_optimizer(self.get_parameters_to_update())
        self.create_loss_calculator()

        self.trainIter  = iter(self.train_loader)

        logging.info("{0:} training start".format(str(datetime.now())[:-7]))
        self.summ_writer = SummaryWriter(self.config['training']['ckpt_save_dir'])
        for it in range(iter_start, iter_max, iter_valid):
            train_scalars = self.training()
            valid_scalars = self.validation()
            glob_it = it + iter_valid
            self.write_scalars(train_scalars, valid_scalars, glob_it)

            if(valid_scalars[metrics] > self.max_val_score):
                self.max_val_score = valid_scalars[metrics]
                self.max_val_it    = glob_it
                self.best_model_wts = copy.deepcopy(self.net.state_dict())
            
            if (glob_it % iter_save ==  0):
                save_dict = {'iteration': glob_it,
                             'valid_pred': valid_scalars[metrics],
                             'model_state_dict': self.net.state_dict(),
                             'optimizer_state_dict': self.optimizer.state_dict()}
                save_name = "{0:}/{1:}_{2:}.pt".format(ckpt_dir, ckpt_prefix, glob_it)
                torch.save(save_dict, save_name) 
                txt_file = open("{0:}/{1:}_latest.txt".format(ckpt_dir, ckpt_prefix), 'wt')
                txt_file.write(str(glob_it))
                txt_file.close()

        # save the best performing checkpoint
        save_dict = {'iteration': self.max_val_it,
                    'valid_pred': self.max_val_score,
                    'model_state_dict': self.best_model_wts,
                    'optimizer_state_dict': self.optimizer.state_dict()}
        save_name = "{0:}/{1:}_{2:}.pt".format(ckpt_dir, ckpt_prefix, self.max_val_it)
        torch.save(save_dict, save_name) 
        txt_file = open("{0:}/{1:}_best.txt".format(ckpt_dir, ckpt_prefix), 'wt')
        txt_file.write(str(self.max_val_it))
        txt_file.close()
        logging.info('The best perfroming iter is {0:}, valid {1:} {2:}'.format(\
            self.max_val_it, metrics, self.max_val_score))
        self.summ_writer.close()


    def infer(self):
        device_ids = self.config['testing']['gpus']
        device = torch.device("cuda:{0:}".format(device_ids[0]))
        self.net.to(device)
        # load network parameters and set the network as evaluation mode
        checkpoint_name = self.get_checkpoint_name()
        checkpoint = torch.load(checkpoint_name, map_location = device)
        self.net.load_state_dict(checkpoint['model_state_dict'])
        
        if(self.config['testing'].get('evaluation_mode', True)):
            self.net.eval()
            
        output_csv   = self.config['testing']['output_csv']
        class_num    = self.config['network']['class_num']
        save_probability = self.config['testing'].get('save_probability', False)
        
        infer_time_list = []
        out_prob_list   = []
        out_lab_list    = []
        with torch.no_grad():
            for data in self.test_loader:
                names  = data['names']
                inputs = self.convert_tensor_type(data['image'])
                inputs = inputs.to(device) 
                
                start_time = time.time()
                out_digit = self.net(inputs)
                infer_time = time.time() - start_time
                infer_time_list.append(infer_time)

                if (self.task_type == "cls"):
                    out_prob  = nn.Softmax(dim = 1)(out_digit).detach().cpu().numpy()
                    out_lab   = np.argmax(out_prob, axis=1)
                else: #self.task_type == "cls_nexcl"
                    out_prob  = nn.Sigmoid()(out_digit).detach().cpu().numpy() 
                    out_lab   = np.asarray(out_prob > 0.5, np.uint8)              
                for i in range(len(names)):
                    print(names[i], out_lab[i])
                    if(self.task_type == "cls"):
                        out_lab_list.append([names[i]] + [out_lab[i]])
                    else:
                        out_lab_list.append([names[i]] + out_lab[i].tolist())
                    out_prob_list.append([names[i]] + out_prob[i].tolist())
        
        with open(output_csv, mode='w') as csv_file:
            csv_writer = csv.writer(csv_file, delimiter=',', 
                                quotechar='"',quoting=csv.QUOTE_MINIMAL)
            head = ['image', 'label']
            if(len(out_lab_list[0]) > 1):
                head = ['image'] + ['label{0:}'.format(i) for i in range(class_num)]
            csv_writer.writerow(head)
            for item in out_lab_list:
                csv_writer.writerow(item)
        if(save_probability):
            prob_csv = output_csv.replace(".csv", "_prob.csv")
            with open(prob_csv, mode='w') as csv_file:
                csv_writer = csv.writer(csv_file, delimiter=',', 
                                    quotechar='"',quoting=csv.QUOTE_MINIMAL)
                head = ['image']+['prob{}'.format(i) for i in range(class_num)]
                csv_writer.writerow(head)
                for item in out_prob_list:
                    csv_writer.writerow(item)

        infer_time_list = np.asarray(infer_time_list)
        time_avg = infer_time_list.mean()
        time_std = infer_time_list.std()
        print("testing time {0:} +/- {1:}".format(time_avg, time_std))




# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn
import numpy as np

def get_soft_label(input_tensor, num_class, data_type = 'float'):
    """
        convert a label tensor to one-hot label 
        input_tensor: tensor with shae [B, 1, D, H, W] or [B, 1, H, W]
        output_tensor: shape [B, num_class, D, H, W] or [B, num_class, H, W]
    """

    shape = input_tensor.shape
    if len(shape) == 5:
        output_tensor = torch.nn.functional.one_hot(input_tensor[:, 0], num_classes = num_class).permute(0, 4, 1, 2, 3)
    elif len(shape) == 4:
        output_tensor = torch.nn.functional.one_hot(input_tensor[:, 0], num_classes = num_class).permute(0, 3, 1, 2)
    else:
        raise ValueError("dimention of data can only be 4 or 5: {0:}".format(len(shape)))
    
    if(data_type == 'float'):
        output_tensor = output_tensor.float()
    elif(data_type == 'double'):
        output_tensor = output_tensor.double()
    else:
        raise ValueError("data type can only be float and double: {0:}".format(data_type))

    return output_tensor

def reshape_tensor_to_2D(x):
    """
    reshape input variables of shape [B, C, D, H, W] to [voxel_n, C]
    """
    tensor_dim = len(x.size())
    num_class  = list(x.size())[1]
    if(tensor_dim == 5):
        x_perm  = x.permute(0, 2, 3, 4, 1)
    elif(tensor_dim == 4):
        x_perm  = x.permute(0, 2, 3, 1)
    else:
        raise ValueError("{0:}D tensor not supported".format(tensor_dim))
    
    y = torch.reshape(x_perm, (-1, num_class)) 
    return y 

def reshape_prediction_and_ground_truth(predict, soft_y):
    """
    reshape input variables of shape [B, C, D, H, W] to [voxel_n, C]
    """
    tensor_dim = len(predict.size())
    num_class  = list(predict.size())[1]
    if(tensor_dim == 5):
        soft_y  = soft_y.permute(0, 2, 3, 4, 1)
        predict = predict.permute(0, 2, 3, 4, 1)
    elif(tensor_dim == 4):
        soft_y  = soft_y.permute(0, 2, 3, 1)
        predict = predict.permute(0, 2, 3, 1)
    else:
        raise ValueError("{0:}D tensor not supported".format(tensor_dim))
    
    predict = torch.reshape(predict, (-1, num_class)) 
    soft_y  = torch.reshape(soft_y,  (-1, num_class))
      
    return predict, soft_y

def get_classwise_dice(predict, soft_y, pix_w = None):
    """
    get dice scores for each class in predict (after softmax) and soft_y
    """
    
    if(pix_w is None):
        y_vol = torch.sum(soft_y,  dim = 0)
        p_vol = torch.sum(predict, dim = 0)
        intersect = torch.sum(soft_y * predict, dim = 0)
    else:
        y_vol = torch.sum(soft_y * pix_w,  dim = 0)
        p_vol = torch.sum(predict * pix_w, dim = 0)
        intersect = torch.sum(soft_y * predict * pix_w, dim = 0)
    dice_score = (2.0 * intersect + 1e-5)/ (y_vol + p_vol + 1e-5)
    return dice_score 



# -*- coding: utf-8 -*-
"""
Caculating the confidence map of labels of training samples,
which is used in the method of SLSR.
    Minqing Zhang et al., Characterizing Label Errors: Confident Learning
    for Noisy-Labeled Image Segmentation, MICCAI 2020.    
"""

from __future__ import print_function, division
import cleanlab
import logging
import os
import scipy
import sys
import torch
import numpy as np 
import pandas as pd
import torch.nn as nn
import torchvision.transforms as transforms
from PIL import Image
from pymic.io.nifty_dataset import NiftyDataset
from pymic.transform.trans_dict import TransformDict
from pymic.util.parse_config import *
from pymic.net_run.agent_seg import SegmentationAgent
from pymic.net_run.infer_func import Inferer

def get_confident_map(gt, pred, CL_type = 'both'):
    """
    gt: ground truth label (one-hot) with shape of NXC
    pred: digit prediction of network with shape of NXC
    """
    prob = scipy.special.softmax(pred, axis = 1)
    if CL_type in ['both', 'Qij']:
        noise = cleanlab.pruning.get_noise_indices(gt, prob, prune_method='both', n_jobs=1)
    elif CL_type == 'Cij':
        noise = cleanlab.pruning.get_noise_indices(gt, pred, prune_method='both', n_jobs=1)
    elif CL_type == 'intersection':
        noise_qij = cleanlab.pruning.get_noise_indices(gt, prob, prune_method='both', n_jobs=1)
        noise_cij = cleanlab.pruning.get_noise_indices(gt, pred, prune_method='both', n_jobs=1)
        noise = noise_qij & noise_cij
    elif CL_type == 'union':
        noise_qij = cleanlab.pruning.get_noise_indices(gt, prob, prune_method='both', n_jobs=1)
        noise_cij = cleanlab.pruning.get_noise_indices(gt, pred, prune_method='both', n_jobs=1)
        noise = noise_qij | noise_cij
    elif CL_type in ['prune_by_class', 'prune_by_noise_rate']:
        noise = cleanlab.pruning.get_noise_indices(gt, prob, prune_method=CL_type, n_jobs=1)
    return noise

class NLLCLSLSR(SegmentationAgent):
    def __init__(self, config, stage = 'test'):
        super(NLLCLSLSR, self).__init__(config, stage)

    def infer_with_cl(self):
        device_ids = self.config['testing']['gpus']
        device = torch.device("cuda:{0:}".format(device_ids[0]))
        self.net.to(device)

        if(self.config['testing'].get('evaluation_mode', True)):
            self.net.eval()
            if(self.config['testing'].get('test_time_dropout', False)):
                def test_time_dropout(m):
                    if(type(m) == nn.Dropout):
                        logging.info('dropout layer')
                        m.train()
                self.net.apply(test_time_dropout)

        ckpt_mode = self.config['testing']['ckpt_mode']
        ckpt_name = self.get_checkpoint_name()
        if(ckpt_mode == 3):
            assert(isinstance(ckpt_name, (tuple, list)))
            self.infer_with_multiple_checkpoints()
            return 
        else:
            if(isinstance(ckpt_name, (tuple, list))):
                raise ValueError("ckpt_mode should be 3 if ckpt_name is a list")

        # load network parameters and set the network as evaluation mode
        checkpoint = torch.load(ckpt_name, map_location = device)
        self.net.load_state_dict(checkpoint['model_state_dict'])

        if(self.inferer is None):
            infer_cfg = self.config['testing']
            class_num = self.config['network']['class_num']
            infer_cfg['class_num'] = class_num 
            self.inferer = Inferer(infer_cfg)
        pred_list  = []
        gt_list    = []
        filename_list = []
        with torch.no_grad():
            for data in self.test_loader:
                images = self.convert_tensor_type(data['image'])
                labels = self.convert_tensor_type(data['label_prob'])
                names  = data['names']
                filename_list.append(names)
                images = images.to(device)
    
                pred = self.inferer.run(self.net, images)
                # convert tensor to numpy
                if(isinstance(pred, (tuple, list))):
                    pred = [item.cpu().numpy() for item in pred]
                else:
                    pred = pred.cpu().numpy()
                data['predict'] = pred
                # inverse transform
                for transform in self.transform_list[::-1]:
                    if (transform.inverse):
                        data = transform.inverse_transform_for_prediction(data) 

                pred = data['predict']
                # conver prediction from N, C, H, W to (N*H*W)*C
                print(names, pred.shape, labels.shape)
                pred_2d = np.swapaxes(pred, 1, 2)
                pred_2d = np.swapaxes(pred_2d, 2, 3)
                pred_2d = pred_2d.reshape(-1, class_num) 
                lab = labels.cpu().numpy()
                lab_2d = np.swapaxes(lab, 1, 2)
                lab_2d = np.swapaxes(lab_2d, 2, 3)
                lab_2d = lab_2d.reshape(-1, class_num) 
                pred_list.append(pred_2d)
                gt_list.append(lab_2d)

        pred_cat = np.concatenate(pred_list)
        gt_cat   = np.concatenate(gt_list)
        gt       = np.argmax(gt_cat, axis = 1)
        gt = gt.reshape(-1).astype(np.uint8)
        print(gt.shape, pred_cat.shape)
        conf = get_confident_map(gt, pred_cat)
        conf = conf.reshape(-1, 256, 256).astype(np.uint8) * 255
        save_dir = self.config['dataset']['root_dir'] + "/slsr_conf"
        for idx in range(len(filename_list)):
            filename = filename_list[idx][0].split('/')[-1]
            conf_map = Image.fromarray(conf[idx])
            dst_path = os.path.join(save_dir, filename)
            conf_map.save(dst_path)

def get_confidence_map():
    if(len(sys.argv) < 2):
        print('Number of arguments should be 3. e.g.')
        print('   python nll_cl.py config.cfg')
        exit()
    cfg_file = str(sys.argv[1])
    config   = parse_config(cfg_file)
    config   = synchronize_config(config)

    # set dataset
    transform_names = config['dataset']['valid_transform']
    transform_list  = []
    transform_dict  = TransformDict
    if(transform_names is None or len(transform_names) == 0):
        data_transform = None 
    else:
        transform_param = config['dataset']
        transform_param['task'] = 'segmentation' 
        for name in transform_names:
            if(name not in transform_dict):
                raise(ValueError("Undefined transform {0:}".format(name))) 
            one_transform = transform_dict[name](transform_param)
            transform_list.append(one_transform)
        data_transform = transforms.Compose(transform_list)
    print('transform list', transform_list)
    csv_file  = config['dataset']['train_csv']
    modal_num = config['dataset'].get('modal_num', 1)
    dataset  = NiftyDataset(root_dir  = config['dataset']['root_dir'],
                            csv_file  = csv_file,
                            modal_num = modal_num,
                            with_label= True,
                            transform = data_transform )

    agent = NLLCLSLSR(config, 'test')
    agent.set_datasets(None, None, dataset)
    agent.transform_list = transform_list
    agent.create_dataset()
    agent.create_network()
    agent.infer_with_cl()

    # create training csv for confidence learning
    df_train = pd.read_csv(csv_file)
    pixel_weight = []
    for i in range(len(df_train["label"])):
        lab_name = df_train["label"][i].split('/')[-1]
        weight_name = "slsr_conf/" + lab_name
        pixel_weight.append(weight_name)
    train_cl_dict = {"image": df_train["image"],
                   "pixel_weight": pixel_weight,
                   "label": df_train["label"]}
    train_cl_csv = csv_file.replace(".csv", "_clslsr.csv")
    df_cl = pd.DataFrame.from_dict(train_cl_dict)
    df_cl.to_csv(train_cl_csv, index = False)

if __name__ == "__main__":
    get_confidence_map()


# -*- coding: utf-8 -*-
"""
An implementation of the 3D U-Net paper:
     Özgün Çiçek, Ahmed Abdulkadir, Soeren S. Lienkamp, Thomas Brox, Olaf Ronneberger:
     3D U-Net: Learning Dense Volumetric Segmentation from Sparse Annotation. 
     MICCAI (2) 2016: 424-432
Note that there are some modifications from the original paper, such as
the use of batch normalization, dropout, and leaky relu here.
"""
from __future__ import print_function, division

import torch
import torch.nn as nn
import numpy as np
from torch.nn.functional import interpolate

class ConvBlock(nn.Module):
    """two convolution layers with batch norm and leaky relu"""
    def __init__(self,in_channels, out_channels, dropout_p):
        """
        dropout_p: probability to be zeroed
        """
        super(ConvBlock, self).__init__()
        self.conv_conv = nn.Sequential(
            nn.Conv3d(in_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm3d(out_channels),
            nn.LeakyReLU(),
            nn.Dropout(dropout_p),
            nn.Conv3d(out_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm3d(out_channels),
            nn.LeakyReLU()
        )
       
    def forward(self, x):
        return self.conv_conv(x)

class DownBlock(nn.Module):
    """Downsampling followed by ConvBlock"""
    def __init__(self, in_channels, out_channels, dropout_p):
        super(DownBlock, self).__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool3d(2),
            ConvBlock(in_channels, out_channels, dropout_p)
        )

    def forward(self, x):
        return self.maxpool_conv(x)

class UpBlock(nn.Module):
    """Upssampling followed by ConvBlock"""
    def __init__(self, in_channels1, in_channels2, out_channels, dropout_p,
                 trilinear=True):
        super(UpBlock, self).__init__()
        self.trilinear = trilinear
        if trilinear:
            self.conv1x1 = nn.Conv3d(in_channels1, in_channels2, kernel_size = 1)
            self.up = nn.Upsample(scale_factor=2, mode='trilinear', align_corners=True)
        else:
            self.up = nn.ConvTranspose3d(in_channels1, in_channels2, kernel_size=2, stride=2)
        self.conv = ConvBlock(in_channels2 * 2, out_channels, dropout_p)

    def forward(self, x1, x2):
        if self.trilinear:
            x1 = self.conv1x1(x1)
        x1 = self.up(x1)
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)



class UNetBlock(nn.Module):
    def __init__(self,in_channels, out_channels, acti_func, acti_func_param):
        super(UNetBlock, self).__init__()
        
        self.in_chns   = in_channels
        self.out_chns  = out_channels
        self.acti_func = acti_func

        self.conv1 = ConvolutionLayer(in_channels,  out_channels, 3, 
                padding = 1, acti_func=get_acti_func(acti_func, acti_func_param))
        self.conv2 = ConvolutionLayer(out_channels, out_channels, 3, 
                padding = 1, acti_func=get_acti_func(acti_func, acti_func_param))

    def forward(self, x):
        x = self.conv1(x)
        x = self.conv2(x)
        return x

class UNet3D(nn.Module):
    def __init__(self, params):
        super(UNet3D, self).__init__()
        self.params    = params
        self.in_chns   = self.params['in_chns']
        self.ft_chns   = self.params['feature_chns']
        self.dropout   = self.params['dropout']
        self.n_class   = self.params['class_num']
        self.trilinear = self.params['trilinear']
        self.deep_sup  = self.params['deep_supervise']
        assert(len(self.ft_chns) == 5 or len(self.ft_chns) == 4)

        self.in_conv= ConvBlock(self.in_chns, self.ft_chns[0], self.dropout[0])
        self.down1  = DownBlock(self.ft_chns[0], self.ft_chns[1], self.dropout[1])
        self.down2  = DownBlock(self.ft_chns[1], self.ft_chns[2], self.dropout[2])
        self.down3  = DownBlock(self.ft_chns[2], self.ft_chns[3], self.dropout[3])
        if(len(self.ft_chns) == 5):
          self.down4  = DownBlock(self.ft_chns[3], self.ft_chns[4], self.dropout[4])
          self.up1 = UpBlock(self.ft_chns[4], self.ft_chns[3], self.ft_chns[3], 
               dropout_p = self.dropout[3], trilinear=self.trilinear) 
        self.up2 = UpBlock(self.ft_chns[3], self.ft_chns[2], self.ft_chns[2], 
               dropout_p = self.dropout[2], trilinear=self.trilinear) 
        self.up3 = UpBlock(self.ft_chns[2], self.ft_chns[1], self.ft_chns[1], 
               dropout_p = self.dropout[1], trilinear=self.trilinear) 
        self.up4 = UpBlock(self.ft_chns[1], self.ft_chns[0], self.ft_chns[0], 
               dropout_p = self.dropout[0], trilinear=self.trilinear) 
    
        self.out_conv = nn.Conv3d(self.ft_chns[0], self.n_class, kernel_size = 1)
        if(self.deep_sup):
            self.out_conv1 = nn.Conv3d(self.ft_chns[1], self.n_class, kernel_size = 1)
            self.out_conv2 = nn.Conv3d(self.ft_chns[2], self.n_class, kernel_size = 1)
            self.out_conv3 = nn.Conv3d(self.ft_chns[3], self.n_class, kernel_size = 1)

    def forward(self, x):
        x0 = self.in_conv(x)
        x1 = self.down1(x0)
        x2 = self.down2(x1)
        x3 = self.down3(x2)
        if(len(self.ft_chns) == 5):
          x4 = self.down4(x3)
          x_d3 = self.up1(x4, x3)
        else:
          x_d3 = x3
        x_d2 = self.up2(x_d3, x2)
        x_d1 = self.up3(x_d2, x1)
        x_d0 = self.up4(x_d1, x0)
        output = self.out_conv(x_d0)
        if(self.deep_sup):
            out_shape = list(output.shape)[2:]
            output1 = self.out_conv1(x_d1)
            output1 = interpolate(output1, out_shape, mode = 'trilinear')
            output2 = self.out_conv2(x_d2)
            output2 = interpolate(output2, out_shape, mode = 'trilinear')
            output3 = self.out_conv3(x_d3)
            output3 = interpolate(output3, out_shape, mode = 'trilinear')
            output = [output, output1, output2, output3]
        return output

if __name__ == "__main__":
    params = {'in_chns':4,
              'class_num': 2,
              'feature_chns':[2, 8, 32, 64],
              'dropout' : [0, 0, 0, 0.5],
              'trilinear': True}
    Net = UNet3D(params)
    Net = Net.double()

    x  = np.random.rand(4, 4, 96, 96, 96)
    xt = torch.from_numpy(x)
    xt = torch.tensor(xt)
    
    y = Net(xt)
    y = y.detach().numpy()
    print(y.shape)



# -*- coding: utf-8 -*-
from __future__ import print_function, division
import logging
import numpy as np
import random
import torch
from torch.optim import lr_scheduler
from pymic.loss.seg.util import get_soft_label
from pymic.loss.seg.util import reshape_prediction_and_ground_truth
from pymic.loss.seg.util import get_classwise_dice
from pymic.loss.seg.dice import DiceLoss
from pymic.net_run_wsl.wsl_abstract import WSLSegAgent
from pymic.util.ramps import get_rampup_ratio

class WSLDMPLS(WSLSegAgent):
    """
    Implementation of the following paper:
        Xiangde Luo, Minhao Hu, Wenjun Liao, Shuwei Zhai, Tao Song, Guotai Wang,
        Shaoting Zhang. ScribblScribble-Supervised Medical Image Segmentation via 
        Dual-Branch Network and Dynamically Mixed Pseudo Labels Supervision.
        MICCAI 2022. 
        https://arxiv.org/abs/2203.02106 
    """
    def __init__(self, config, stage = 'train'):
        net_type = config['network']['net_type']
        if net_type not in ['UNet2D_DualBranch', 'UNet3D_DualBranch']:
            raise ValueError("""For WSL_DMPLS, a dual branch network is expected. \
                It only supports UNet2D_DualBranch and UNet3D_DualBranch currently.""")
        super(WSLDMPLS, self).__init__(config, stage)

    def training(self):
        class_num   = self.config['network']['class_num']
        iter_valid  = self.config['training']['iter_valid']
        wsl_cfg     = self.config['weakly_supervised_learning']
        iter_max     = self.config['training']['iter_max']
        rampup_start = wsl_cfg.get('rampup_start', 0)
        rampup_end   = wsl_cfg.get('rampup_end', iter_max)
        train_loss  = 0
        train_loss_sup = 0
        train_loss_reg = 0
        train_dice_list = []
        self.net.train()
        for it in range(iter_valid):
            try:
                data = next(self.trainIter)
            except StopIteration:
                self.trainIter = iter(self.train_loader)
                data = next(self.trainIter)
            
            # get the inputs
            inputs = self.convert_tensor_type(data['image'])
            y      = self.convert_tensor_type(data['label_prob'])  
                         
            inputs, y = inputs.to(self.device), y.to(self.device)
            
            # zero the parameter gradients
            self.optimizer.zero_grad()
                
            # forward + backward + optimize
            outputs1, outputs2 = self.net(inputs)
            loss_sup1 = self.get_loss_value(data, outputs1, y) 
            loss_sup2 = self.get_loss_value(data, outputs2, y) 
            loss_sup  = 0.5 * (loss_sup1 + loss_sup2)

            # get pseudo label with dynamical mix
            outputs_soft1 = torch.softmax(outputs1, dim=1)
            outputs_soft2 = torch.softmax(outputs2, dim=1)
            beta = random.random()
            pseudo_lab = beta*outputs_soft1.detach() + (1.0-beta)*outputs_soft2.detach()
            pseudo_lab = torch.argmax(pseudo_lab, dim = 1, keepdim = True)
            pseudo_lab = get_soft_label(pseudo_lab, class_num, self.tensor_type)
            
            # calculate the pseudo label supervision loss
            loss_calculator = DiceLoss()
            loss_dict1 = {"prediction":outputs1, 'ground_truth':pseudo_lab}
            loss_dict2 = {"prediction":outputs2, 'ground_truth':pseudo_lab}
            loss_reg   = 0.5 * (loss_calculator(loss_dict1) + loss_calculator(loss_dict2))
            
            rampup_ratio = get_rampup_ratio(self.glob_it, rampup_start, rampup_end, "sigmoid")
            regular_w = wsl_cfg.get('regularize_w', 0.1) * rampup_ratio
            loss = loss_sup + regular_w*loss_reg

            loss.backward()
            self.optimizer.step()
            if(self.scheduler is not None and \
                not isinstance(self.scheduler, lr_scheduler.ReduceLROnPlateau)):
                self.scheduler.step()

            train_loss = train_loss + loss.item()
            train_loss_sup = train_loss_sup + loss_sup.item()
            train_loss_reg = train_loss_reg + loss_reg.item() 
            # get dice evaluation for each class in annotated images
            if(isinstance(outputs1, tuple) or isinstance(outputs1, list)):
                outputs1 = outputs1[0] 
            p_argmax = torch.argmax(outputs1, dim = 1, keepdim = True)
            p_soft   = get_soft_label(p_argmax, class_num, self.tensor_type)
            p_soft, y = reshape_prediction_and_ground_truth(p_soft, y) 
            dice_list   = get_classwise_dice(p_soft, y)
            train_dice_list.append(dice_list.cpu().numpy())
        train_avg_loss = train_loss / iter_valid
        train_avg_loss_sup = train_loss_sup / iter_valid
        train_avg_loss_reg = train_loss_reg / iter_valid
        train_cls_dice = np.asarray(train_dice_list).mean(axis = 0)
        train_avg_dice = train_cls_dice.mean()

        train_scalers = {'loss': train_avg_loss, 'loss_sup':train_avg_loss_sup,
            'loss_reg':train_avg_loss_reg, 'regular_w':regular_w,
            'avg_dice':train_avg_dice,     'class_dice': train_cls_dice}
        return train_scalers
        